from PyQt5 import QtWidgets, QtCore
from PyQt5.QtWidgets import QPushButton, QComboBox, QApplication, QMainWindow, QTabWidget, QWidget, QVBoxLayout, QLabel, QLineEdit, QSpinBox, QCheckBox
from PyQt5.QtCore import QObject, pyqtSignal, QThread, Qt
from PyQt5.QtGui import QPixmap, QIcon, QPainter, QColor, QFont, QIntValidator
from PyQt5.QtWidgets import QFrame, QTextEdit, QScrollArea, QSplashScreen, QGroupBox, QApplication, QTabWidget, QFileDialog, QWidget, QVBoxLayout, QLabel, QLineEdit, QGridLayout, QPushButton, QCheckBox, QSpinBox
import sys
import json
import subprocess
from pathlib import Path
from lxml import etree


#TODO: Replace default values with setPlaceholderText()


current_dir = Path(sys.argv[0]).resolve().parent

#=====================
#Worker Threads
#=====================
class MosaicWorkerThread(QThread):
    #output_signal for output_text box
    output_signal = pyqtSignal(str)

    def __init__(self, command):
        super(MosaicWorkerThread, self).__init__()
        self.command = command

    def run(self):
        process = subprocess.Popen(self.command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, text=True, bufsize=1)
        with process.stdout:
            #Read each line of the process output as it is generated
            for line in iter(process.stdout.readline, ''):
                #Emit the line to the output_text box
                self.output_signal.emit(line)
        process.wait()
        #Check the return code of the process
        if process.returncode != 0:
            self.output_signal.emit(f"Process finished with error code {process.returncode}")
        else:
            self.output_signal.emit("Mosaic script finished successfully!")
class BatchThread(QThread):
    #finished_signal for output_text_mosaic box
    finished_signal = pyqtSignal(str)

    def __init__(self, batch_command):
        super(BatchThread, self).__init__()
        self.batch_command = batch_command

    def run(self):
        # Directly execute the batch file without 'cmd /c start'
        process = subprocess.Popen(self.batch_command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True,
                                   text=True, bufsize=1)
        # Wait for the process to finish and capture the output
        stdout, stderr = process.communicate()
        # Check the return code of the process
        if process.returncode == 0:
            self.finished_signal.emit(f"Batch {self.batch_command} completed successfully.")
        else:
            self.finished_signal.emit(f"Error in batch {self.batch_command}: {stderr.decode()}")
class WorkerThread(QThread):
    #output_signal captures output for output_text box
    #batch_files_signal captures batch files
    #output_folder_signal captures output folders
    output_signal = pyqtSignal(str)
    batch_files_signal = pyqtSignal(list)
    output_folder_signal = pyqtSignal(str)

    def __init__(self, command):
        super(WorkerThread, self).__init__()
        self.command = command

    def run(self):
        self.process = subprocess.Popen(self.command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, text=True, bufsize=1)
        #batch_files stores the batch files when flag_NumberofBatches >1
        batch_files = []
        #output_folders stores the output folders - formatted as a list to account for alternate scenarios
        output_folders = []
        with self.process.stdout:
            #Read each line of the process output as it is generated
            for line in iter(self.process.stdout.readline, ''):
                #If the line ends with '.bat', add it to the batch_files list and do not emit it to the output_text box
                if line.strip().endswith('.bat'):
                    batch_files.append(line.strip())
                #If the line starts with 'output_folder =', add it to the output_folders list and do not emit it to the output_text box
                elif line.startswith('output_folder ='):
                    output_folder = line.split('=')[1].strip()
                    output_folders.append(output_folder)
                #Else, emit the line to the output_text box
                else:
                    self.output_signal.emit(line)
        self.process.wait()
        #Check the return code of the process
        if self.process.returncode != 0:
            self.output_signal.emit(f"Process finished with error code {self.process.returncode}")
        else:
            #Emit the batch files and output folders to the main thread
            self.batch_files_signal.emit(batch_files)
            for output_folder in output_folders:
                self.output_folder_signal.emit(output_folder)
            #Emit a success message to the output_text box
            self.output_signal.emit("i-Tree Cool Air Prep script finished successfully!\nSelect 'Run Cool Air' or navigate to your project folder's .bat files to run the Cool Air model.")
    def terminate(self):
        if self.process and self.process.poll() is None:
            self.process.terminate()
            self.process.wait()
#=====================
#Loading Screen
#=====================
class SplashScreen(QSplashScreen):
    def __init__(self):
        # Load the original logo image
        pix_map = current_dir / 'resources' / 'iTreeLogo.png'
        logoPixmap = QPixmap(str(pix_map))
        #Create larger pixel map with margins to display text
        largerPixmap = QPixmap(logoPixmap.width() + 60, logoPixmap.height() + 90)
        # Fill the pixmap with a white background
        largerPixmap.fill(QColor(Qt.white))

        # Draw the logo onto the center of the larger pixmap
        painter = QPainter(largerPixmap)
        painter.drawPixmap(30, 30, logoPixmap)
        painter.end()
        #Set window icon
        super().__init__(largerPixmap)
        self.setWindowIcon(QIcon(str(pix_map)))

        # Loading message
        self.setFont(QFont('Arial', 12, QFont.Bold))
        self.showMessage("i-Tree Research Suite - Cool Air\nLoading...",
                         int(QtCore.Qt.AlignBottom | QtCore.Qt.AlignCenter), QtCore.Qt.black)
#=====================
#Help Windows
#=====================
#Main
class HelpWindow(QWidget):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.initUI()

    def initUI(self):
        self.setWindowTitle('Help Window')
        icon_path = current_dir / 'resources' / 'iTree_transparent.ico'
        #Set window icon
        self.setWindowIcon(QIcon(str(icon_path)))
        layout = QtWidgets.QVBoxLayout()
        #Add help text for each section
        #Help Text - Project Details
        project_details_help = QLabel("Project Details")
        project_details_help.setStyleSheet("font-weight: bold;")
        project_details_help_window = QtWidgets.QTextEdit()
        project_details_help_window.setReadOnly(True)
        project_details_help_text = QLabel("Project Name: Name of the project\n\n"
                                           "Columns/Rows Max: Maximum columns and rows, up to 300\n\n"
                                           "Simulation Start/End: Start and end dates of the simulation (YYYYMMDD or YYYYMMDDHH)\n\n"                                           
                                           "Output Resolution: Desired resolution of the output, comma-separated for multiples (ex: 30, 90)\n\n"
                                           "Output EPSG: EPSG code for the output coordinate system (ex: 5070)")
        project_details_help_window.append(project_details_help_text.text())
        #Help Text - Weather
        weather_help = QLabel("Weather")
        weather_help.setStyleSheet("font-weight: bold;")
        weather_help_window = QtWidgets.QTextEdit()
        weather_help_window.setReadOnly(True)
        weather_help_text = QLabel("Multiple Weather Stations: Use multiple weather stations (Requires multiple station CSV path)\n\n"
                                   "Hottest Day: Calculate hottest day for simulation - overrides Simulation Start/End dates\n\n"
                                    "Weather Path: Path to folder containing Weather and Radiation.csv for single weather stations\n\n" 
                                    "Multiple Station CSV: Path to the multiple station file")
        weather_help_window.append(weather_help_text.text())
        #Help Text - Project Paths
        paths_help = QLabel("Paths")
        paths_help.setStyleSheet("font-weight: bold;")
        paths_help_window = QtWidgets.QTextEdit()
        paths_help_window.setReadOnly(True)
        paths_help_text = QLabel(   "Project Set-Up Folder: Path to the trunk folder\n\n"
                                    "Project Folder: Path to the project folder\n\n"
                                    "AOI Filepath: Path to the AOI shapefile\n\n"
                                    "DEM Filepath: Path to the DEM file\n\n"
                                    "DEM Units: Units of the DEM file\n\n"
                                    "Land Cover Folder: Path to the folder containing impervious, tree, and land cover files\n\n"
                                    "NLCD Auto Search: Automatically search for NLCD file\n\n"
                                    "NLCD Auto Search Year: Year for NLCD auto search\n\n"
                                    "[Optional] Anthropogenic Heat: Path to the anthropogenic heat raster \n\n"
                                    "[Optional] Census Filepath: Path to polygon containing sub-sets of AOI\n\n"
                                    "[Optional] Census FID List: List of FIDs to process\n\n")
        paths_help_window.append(paths_help_text.text())

        project_details_help_window.setMinimumHeight(150)
        weather_help_window.setMinimumHeight(125)
        paths_help_window.setMinimumHeight(250)
        layout.addWidget(project_details_help)
        layout.addWidget(project_details_help_window)
        layout.addWidget(weather_help)
        layout.addWidget(weather_help_window)
        layout.addWidget(paths_help)
        layout.addWidget(paths_help_window)
        self.setLayout(layout)
#Advanced Config
class HelpWindow_Adv(QWidget):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.initUI()

    def initUI(self):
        self.setWindowTitle('Advanced Configuration: Help')
        icon_path = current_dir / 'resources' / 'iTree_transparent.ico'
        self.setWindowIcon(QIcon(str(icon_path)))
        layout = QtWidgets.QVBoxLayout()
        #Help Text - Advanced Config
        advanced_config_help = QLabel("Advanced Configuration")
        advanced_config_help.setStyleSheet("font-weight: bold;")
        #Add text box for help text
        advanced_config_help_window = QtWidgets.QTextEdit()
        advanced_config_help_window.setReadOnly(True)
        advanced_config_help_text = QLabel("Calculate RefWeatherLocationElevation_m: Calculate average elevation for use in TemperatureExecutionParams; represents elevation of weather station\n\n"
                                           "Parallel batches: Set amount of batches to run in parallel for Cool Air. Recommended batches based on RAM:\n"
                                           "<=8 GB: Up to 4 batches\n"
                                           "8-16 GB: Up to 8 batches\n"
                                           "16-32 GB: Up to 10 batches\n\n"
                                           "Preset Soil Type: Set soil type to update HydroPlusConfig.xml elements\n\n")
        advanced_config_help_window.append(advanced_config_help_text.text())

        hydro_calibration_help = QLabel("Hydro Calibration")
        hydro_calibration_help.setStyleSheet("font-weight: bold;")
        hydro_calibration_help_window = QtWidgets.QTextEdit()
        hydro_calibration_help_window.setReadOnly(True)
        hydro_calibration_help_text = QLabel("This window allows for set-up of Hydro calibration. It is recommended to always perform a Hydro calibration before running i-Tree Cool Air, as this will "
                                             "improve the accuracy of the temperature outputs.\n\n"
                                             "Generate Hydro Inputs: Checkbox enabling Hydro Inputs being generated\n\n"
                                             "Calibration: Checkbox enabling calibration (recommended)\n\n"
                                             "Auto-Calculate Vadose Zone: Automatically calculate start and end vadose zone (used for PEST calibration) using an initial run and extracting the mean vadose depth. It is recommended to keep this enabled.\n\n"
                                             "Start Vadose Zone: Start vadose zone depth (mm)\n\n"
                                             "End Vadose Zone: End vadose zone depth (mm)\n\n"
                                             "Calibration Start: Start date for calibration (YYYYMMDD). Typically set to YYYY1001 for start of 'water year'.\n\n"
                                             "Calibration End: End date for calibration (YYYYMMDD). Typically set to YYYY0930 for end of 'water year'. An optimal calibration will be at least two years long, and representing near-average precipitation for the area.\n\n")
        hydro_calibration_help_window.append(hydro_calibration_help_text.text())
        #Help Text - XML Editor
        xml_help = QLabel("XML Editor")
        xml_help.setStyleSheet("font-weight: bold;")
        xml_help_window = QtWidgets.QTextEdit()
        xml_help_window.setReadOnly(True)
        xml_help_text = QLabel("The XML Editor allows editing of the HydroPlusConfig_CoolAir_Template.xml file.\n\n"
                               "If 'Import XML' is checked and a HydroPlusConfig file is found, no values will be taken from the XML Editor.\n\n")
        xml_help_window.append(xml_help_text.text())


        layout.addWidget(advanced_config_help)
        layout.addWidget(advanced_config_help_window)
        layout.addWidget(hydro_calibration_help)
        layout.addWidget(hydro_calibration_help_window)
        layout.addWidget(xml_help)
        layout.addWidget(xml_help_window)
        self.setLayout(layout)
#Scenarios
class HelpWindow_Sce(QWidget):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.initUI()

    def initUI(self):
        self.setWindowTitle('Help Window')
        icon_path = current_dir / 'resources' / 'iTree_transparent.ico'
        #Set window icon
        self.setWindowIcon(QIcon(str(icon_path)))
        layout = QtWidgets.QVBoxLayout()
        #Add help text for each section
        #Help Text - Alternate Scenarios Overview
        scenario_help = QLabel("Alternate Scenario Set-Up")
        scenario_help.setStyleSheet("font-weight: bold;")
        scenario_help_window = QtWidgets.QTextEdit()
        scenario_help_window.setReadOnly(True)
        scenario_help_text = QLabel("Alternate scenarios allow for comparison of multiple Cool Air simulations, with adjustments made based on the land, imperivous, and/or tree cover file.\n\n"
                                    "The 'Where' section allows for filtering based on the land, imperivous, and/or tree cover file. Currently, only one raster can be used the in the 'where' statement."
                                    "Impervious and tree cover files should be formatted between 0-100%, whereas land cover follows NLCD land cover classifications.\n\n"
                                    "The 'Action' section defines what changes will be made based on the filtered area.\n")
        scenario_help_window.append(scenario_help_text.text())
        layout.addWidget(scenario_help)
        layout.addWidget(scenario_help_window)

        # Help Text - Common Alternate Scenarios
        common_scenario_help = QLabel("Common Alternate Scenarios")
        common_scenario_help.setStyleSheet("font-weight: bold;")
        common_scenario_window = QtWidgets.QTextEdit()
        common_scenario_window.setReadOnly(True)
        common_scenario_text = QLabel(
            "TC10: Increase tree cover by 10% across the entire simulation area, excluding water and wetlands. \n\n"
            "   WHERE: land cover is greater than 11, AND land cover is less than 90\n"
            "   ACTION: add 10 for raster: tree cover\n\n"
            "TC10_U: Increase tree cover by 10% in urban areas only.\n\n"
            "   WHERE land cover is greater than or equal to 21 AND land cover is less than or equal to 24\n"
            "   ACTION: add 10 for raster: tree cover\n\n"
            "TC_30_Minimum: Increase tree cover to 30% in areas with less than 30% tree cover.\n\n"
            "   WHERE: tree cover is less than 30\n"
            "   ACTION: set equal to 30 for raster: tree cover\n")
        common_scenario_window.append(common_scenario_text.text())
        layout.addWidget(common_scenario_help)
        layout.addWidget(common_scenario_window)


        tips_label = QLabel("Tips")
        tips_label.setStyleSheet("font-weight: bold;")
        tips_window = QtWidgets.QTextEdit()
        tips_window.setReadOnly(True)
        tips_text = QLabel("1. Use the '+' button to add additional 'Where' or 'Action' fields.\n"
                            "2. Use the 'X' button to remove the last 'Where' or 'Action' fields.\n"
                            "3. Make sure the Scenario Name does not contain any spaces; separate words with underscores.\n"
                            "4. Only use numbers for the fill-in values.\n"
                            "4. Scenarios are saved to the JSON Config file for your project, and can be reloaded for future use.")
        tips_window.append(tips_text.text())
        layout.addWidget(tips_label)
        layout.addWidget(tips_window)
        #Adjust layout to make tips the smallest section
        scenario_help_window.setMinimumHeight(150)
        common_scenario_window.setMinimumHeight(250)
        tips_window.setMinimumHeight(10)

        self.setLayout(layout)
#=====================
#Main Window
#=====================
class MainWindow(QMainWindow):
    #~~~~~~~~~~~~~
    # SIGNALS
    #~~~~~~~~~~~~~
    #If multiple weather stations is checked, enabled hottest day and multi-station input
    def handle_multiple_weather_toggle(self, checked):
        if checked:
            self.hottest_day_checkbox.setEnabled(True)
            self.multi_station_input.setEnabled(True)
            self.weather_path_input.setEnabled(False)
        else:
            self.hottest_day_checkbox.setChecked(False)
            self.hottest_day_checkbox.setEnabled(False)
            self.multi_station_input.setEnabled(False)
            self.weather_path_input.setEnabled(True)
    #If NLCD autosearch is checked, disable manual NLCD inputs and enable year input
    def handle_nlcd_autosearch_toggle(self, checked):
        if checked:
            self.nlcd_land_input.setEnabled(False)
            self.nlcd_imperv_input.setEnabled(False)
            self.nlcd_tree_input.setEnabled(False)
            self.nlcd_auto_search_year.setEnabled(True)
        else:
            self.nlcd_land_input.setEnabled(True)
            self.nlcd_imperv_input.setEnabled(True)
            self.nlcd_tree_input.setEnabled(True)
            self.nlcd_auto_search_year.setEnabled(False)
    #If batches are received from p_itca_01, enable 'Run Cool Air' button and 'Auto-Run Mosaic' button.
    def handle_batch_toggle(self, batch_files):
        if batch_files:
            self.batch_files = batch_files
            self.run_button.setEnabled(True)
            self.auto_run_mosaic_button.setEnabled(True)
    #If census path is filled in, allow the fill-in of the census FIDs for processing
    def update_census_list_toggle(self):
        # Check if the input is not empty and appears to be a valid file path
        census_path = self.census_path_input.text().strip()
        self.census_list_input.setEnabled(bool(census_path))  # Enable if census_path is non-empty
        #Update
        self.line_edits_exec['PrintBlockGroupDailyAndHourly'].setText('True')

    def update_census_list(self):
        census_fid_list = self.census_list_input.text().strip()
        self.line_edits_exec['PrintBlockGroupRange0'].setText(census_fid_list)

    # Function to enable or disable hydro fields
    def toggle_hydro_fields(self):
        enabled = self.hydro_prep_checkbox.isChecked()
        self.calibration_checkbox.setEnabled(enabled)
        self.vadose_zone_checkbox.setEnabled(enabled)
        self.start_vadose_zone_label.setEnabled(enabled and not self.vadose_zone_checkbox.isChecked())
        self.start_vadose_zone_input.setEnabled(enabled and not self.vadose_zone_checkbox.isChecked())
        self.end_vadose_zone_label.setEnabled(enabled and not self.vadose_zone_checkbox.isChecked())
        self.end_vadose_zone_input.setEnabled(enabled and not self.vadose_zone_checkbox.isChecked())
        self.calibration_start_label.setEnabled(enabled and self.calibration_checkbox.isChecked())
        self.calibration_start_input.setEnabled(enabled and self.calibration_checkbox.isChecked())
        self.calibration_end_label.setEnabled(enabled and self.calibration_checkbox.isChecked())
        self.calibration_end_input.setEnabled(enabled and self.calibration_checkbox.isChecked())
    # Function to toggle vadose zone fields
    def toggle_vadose_zone_fields(self):
        enabled = not self.vadose_zone_checkbox.isChecked()
        self.start_vadose_zone_label.setEnabled(enabled)
        self.start_vadose_zone_input.setEnabled(enabled)
        self.end_vadose_zone_label.setEnabled(enabled)
        self.end_vadose_zone_input.setEnabled(enabled)
    # Function to toggle calibration fields
    def toggle_calibration_fields(self):
        enabled = self.calibration_checkbox.isChecked()
        self.calibration_start_label.setEnabled(enabled)
        self.calibration_start_input.setEnabled(enabled)
        self.calibration_end_label.setEnabled(enabled)
        self.calibration_end_input.setEnabled(enabled)

    #~~~~~~~~~~~~~
    # JSON IMPORT
    #~~~~~~~~~~~~~
    #Load scenarios from JSON file, called in import_json
    def load_scenarios(self, data):
        all_scenarios = data.get('all_scenarios_data', [])
        for i, scenario_data in enumerate(all_scenarios):
            # Ensure enough scenario boxes exist
            # Currently only configured for 3 scenarios
            if i < len(self.scenario_boxes):
                scenario_box = self.scenario_boxes[i]
                #print(f"Scenario {i + 1}: scenario_box found.")

                # Retrieve alias input
                alias_input = scenario_box.findChild(QLineEdit)
                if alias_input:
                    alias_name = scenario_data.get("name", f"Alt Case {i + 1}")
                    alias_input.setText(alias_name)
                    #print(f"Set alias input to: {alias_name}")

                # Retrieve and set WHERE fields
                where_data = scenario_data.get("where", [])
                for group_box in scenario_box.findChildren(QGroupBox):
                    if "WHERE" in group_box.title():
                        where_group_box = group_box
                        break
                else:
                    where_group_box = None

                if where_group_box:
                    #print(f"Scenario {i + 1}: where_group_box found.")
                    where_layout = where_group_box.layout()
                    for row, where_item in enumerate(where_data):
                        #print(f"Processing WHERE row {row + 1} for Scenario {i + 1}")

                        # Ensure enough rows exist in the layout, add new row if needed
                        if row >= where_layout.rowCount():
                            self.add_where_row(where_layout, where_item.get("field", ""))
                        # Retrieve widgets for each field in the layout
                        #AND/OR
                        operator_widget = where_layout.itemAtPosition(row, 0).widget()
                        #Field - land cover, tree cover, impervious cover, DEM
                        field_widget = where_layout.itemAtPosition(row, 1).widget()
                        #Condition - equals, greater than, less than, etc.
                        condition_widget = where_layout.itemAtPosition(row, 2).widget()
                        #Value - user input; integer or float
                        value_widget = where_layout.itemAtPosition(row, 3).widget()
                        #Set each field's value based on the JSON data
                        if isinstance(operator_widget, QComboBox):
                            operator = where_item.get("operator", "")
                            operator_widget.setCurrentText(operator)
                            #print(f"Set WHERE operator to: {operator}")
                        if isinstance(field_widget, QLabel):
                            field = where_item.get("field", "")
                            field_widget.setText(field)
                            #print(f"Set WHERE field to: {field}")
                        if isinstance(condition_widget, QComboBox):
                            condition = where_item.get("condition", "")
                            condition_widget.setCurrentText(condition)
                            #print(f"Set WHERE condition to: {condition}")
                        if isinstance(value_widget, QLineEdit):
                            value = where_item.get("value", "")
                            value_widget.setText(value)
                            #print(f"Set WHERE value to: {value}")

                # Retrieve and set ACTION fields
                action_data = scenario_data.get("action", [])
                for group_box in scenario_box.findChildren(QGroupBox):
                    if "ACTION" in group_box.title():
                        action_group_box = group_box
                        break
                else:
                    action_group_box = None

                if action_group_box:
                    action_layout = action_group_box.layout()
                    for row, action_item in enumerate(action_data):
                        #print(f"Processing ACTION row {row + 1} for Scenario {i + 1}")
                        # Ensure enough rows exist in the layout, add new row if needed
                        if row >= action_layout.rowCount():
                            self.add_action_row(action_layout)
                        # Retrieve widgets for each field in the layout
                        operation_widget = action_layout.itemAtPosition(row, 1).widget()
                        value_widget = action_layout.itemAtPosition(row, 2).widget()
                        field_widget = action_layout.itemAtPosition(row, 4).widget()
                        #Set each field's value based on the JSON data
                        if isinstance(operation_widget, QComboBox):
                            operation = action_item.get("operation", "")
                            operation_widget.setCurrentText(operation)
                            #print(f"Set ACTION operation to: {operation}")
                        if isinstance(value_widget, QLineEdit):
                            value = action_item.get("value", "")
                            value_widget.setText(value)
                            #print(f"Set ACTION value to: {value}")
                        if isinstance(field_widget, QComboBox):
                            field = action_item.get("field", "")
                            field_widget.setCurrentText(field)
                            #print(f"Set ACTION field to: {field}")
    #Set XML element text based on JSON data
    def set_xml_element(self, element_dict, key, value):
        if key in element_dict:
            element_dict[key].setText(value)
    #Import JSON file
    def import_json(self):
        # Open file dialog to select a JSON file
        json_file, _ = QFileDialog.getOpenFileName(self, "Select JSON file", "",
                                                   "JSON Files (*.json);;All Files (*)")
        if json_file:
            # Read the JSON file
            with open(json_file, 'r') as file:
                data = json.load(file)

            # Update the text inputs in the GUI from the JSON data
            #Project Details
            self.project_name_input.setText(data.get('proj_name', ''))
            output_resolution_list = data.get('output_resolution_list', [])
            # If output_resolution_list is a list, join the items with a comma
            if isinstance(output_resolution_list, list):
                output_resolution_str = ', '.join(output_resolution_list)  # Join list items with a comma
            else:
                output_resolution_str = str(output_resolution_list)
            self.output_resolution_input.setText(output_resolution_str)
            self.simulation_start_input.setText(data.get('simulation_start', ''))
            self.simulation_end_input.setText(data.get('simulation_end', ''))
            self.col_row_max_input.setText(str(data.get('col_max', '')))
            self.output_epsg_input.setText(data.get('flag_OutputEPSG', ''))
            #Weather
            flag_multiple_station = data.get('flag_MultipleStation', '')
            if flag_multiple_station:
                self.multiple_weather_checkbox.setChecked(True)
                self.multi_station_input.setText(data.get('filepath_multistation', ''))
            else:
                self.weather_path_input.setText(data.get('path_weather', ''))
            flag_hottest_day_simulated = data.get('flag_HottestDaySimulated', '')
            if flag_hottest_day_simulated:
                self.hottest_day_checkbox.setChecked(True)
            #Paths
            self.trunk_path_input.setText(data.get('path_trunk', ''))
            self.project_path_input.setText(data.get('path_project', ''))
            self.aoi_path_input.setText(data.get('filepath_aoi', ''))
            self.dem_path_input.setText(data.get('filepath_dem', ''))
            #Set DEM units based on flag_DEM_ratio
            dem_ratio = data.get('flag_DEM_ratio', '')
            if dem_ratio == 100.0:
                self.dem_units_dropdown.setCurrentText('cm')
            elif dem_ratio == 1.0:
                self.dem_units_dropdown.setCurrentText('m')

            self.nlcd_path_input.setText(data.get('path_nlcd', ''))
            flag_nlcd_auto_search = data.get('flag_NLCDAutoSearch', '')
            if flag_nlcd_auto_search:
                self.nlcd_auto_search.setChecked(True)
                self.nlcd_auto_search_year.setValue(int(data.get('flag_NLCD_LC_year', '')))
            else:
                self.nlcd_land_input.setText(data.get('file_lc', ''))
                self.nlcd_imperv_input.setText(data.get('file_ic', ''))
                self.nlcd_tree_input.setText(data.get('file_tc', ''))

            flag_anthropogenic_heat = data.get('flag_AnthropogenicHeat', '')
            if flag_anthropogenic_heat:
                self.ah_path_input.setText(data.get('filepath_ah4guc', ''))
            flag_census = data.get('flag_Census', '')
            if flag_census:
                self.census_path_input.setText(data.get('filepath_census', ''))
                self.census_list_input.setText(data.get('census_fids', ''))
            #Advanced Config
            flag_xml_import = data.get('flag_XML_Import', '')
            if flag_xml_import:
                self.xml_import_checkbox.setChecked(True)
            else:
                self.xml_import_checkbox.setChecked(False)
            flag_dem_avg = data.get('flag_DEM_average', '')
            if flag_dem_avg:
                self.dem_avg_checkbox.setChecked(True)
            soil_type = data.get('soil_type', '')
            if soil_type:
                self.soil_type_dropdown.setCurrentText(soil_type)
            flag_hydro = data.get('flag_Hydro', '')
            if flag_hydro:
                self.hydro_prep_checkbox.setChecked(True)
            self.batch_number_spinbox.setValue(data.get('flag_NumberofBatches', ''))
            self.batch_number_spinbox.setValue(data.get('flag_NumberofBatches', ''))
            flag_calibration = data.get('flag_Calibration', '')
            if flag_calibration:
                self.calibration_checkbox.setChecked(True)
                self.calibration_start_input.setText(data.get('calibration_start', ''))
                self.calibration_end_input.setText(data.get('calibration_end', ''))
                flag_Vadose_auto = data.get('flag_Vadose_auto', '')
                if flag_Vadose_auto:
                    self.vadose_zone_checkbox.setChecked(True)
                else:
                    self.start_vadose_zone_input.setText(data.get('first_vadose', ''))
                    self.end_vadose_zone_input.setText(data.get('last_vadose', ''))


            # XML Editor
            data_folder_params = data.get('xml_import', {}).get('DataFolder', {})
            for key, value in data_folder_params.items():
                self.set_xml_element(self.line_edits, key, value)

            temp_calc_params = data.get('xml_import', {}).get('TemperatureCalculationParams', {})
            for key, value in temp_calc_params.items():
                self.set_xml_element(self.line_edits_temp, key, value)

            temp_exec_params = data.get('xml_import', {}).get('TemperatureExecutionParams', {})
            for key, value in temp_exec_params.items():
                self.set_xml_element(self.line_edits_exec, key, value)

            #Scenarios
            self.load_scenarios(data)
    #~~~~~~~~~~~~~
    # SCENARIO LOGIC
    #~~~~~~~~~~~~~
    #Adds first 'Where' and 'Action' fields for each scenario
    def add_where_fields(self, layout):
        # Add "Where" section
        where_label = QLabel("")
        layout.addWidget(where_label, 0, 0)

        where_field = QComboBox()
        where_field.addItems(["land cover", "tree cover", "impervious cover", "DEM"])
        layout.addWidget(where_field, 0, 1)

        where_condition = QComboBox()
        where_condition.addItems(["equals", "is greater than", "is less than",
                                  "is greater than or equal to", "is less than or equal to",
                                  "is not equal to"])
        layout.addWidget(where_condition, 0, 2)

        where_value = QLineEdit()
        layout.addWidget(where_value, 0, 3)

        add_where_button = QPushButton("+")
        add_where_button.setMaximumWidth(30)
        add_where_button.clicked.connect(lambda: self.add_where_row(layout, where_field.currentText()))
        layout.addWidget(add_where_button, 0, 4)

        # Store initial field type for adding new rows
        self.initial_where_field_type = where_field.currentText()
    def add_action_fields(self, layout):
        # Add "Action" section
        action_label = QLabel("")
        layout.addWidget(action_label, 0, 0)

        action_operation = QComboBox()
        action_operation.addItems(["add", "subtract", "multiply", "divide", "set equal to"])
        layout.addWidget(action_operation, 0, 1)

        action_value = QLineEdit()
        layout.addWidget(action_value, 0, 2)

        action_filetype_label = QLabel("FOR RASTER:")
        layout.addWidget(action_filetype_label, 0, 3)

        action_field = QComboBox()
        action_field.addItems(["land cover", "tree cover", "impervious cover", "DEM"])
        layout.addWidget(action_field, 0, 4)

        add_action_button = QPushButton("+")
        add_action_button.setMaximumWidth(30)
        add_action_button.clicked.connect(lambda: self.add_action_row(layout))
        layout.addWidget(add_action_button, 0, 5)
    #Adds new 'Where' and 'Action' fields for each scenario
    #Where fields are added based on the initial field type and can not be changed
    #TODO: Figure out how to run scenarios referencing multiple rasters
    def add_where_row(self, layout, initial_field_type):
        # Get the current number of rows in the layout
        row_position = layout.rowCount()

        and_or_button = QComboBox()
        and_or_button.addItems(["AND", "OR"])
        layout.addWidget(and_or_button, row_position, 0)

        where_field = QLabel(initial_field_type)
        layout.addWidget(where_field, row_position, 1)

        where_condition = QComboBox()
        where_condition.addItems(["equals", "is greater than", "is less than",
                                  "is greater than or equal to", "is less than or equal to",
                                  "is not equal to"])
        layout.addWidget(where_condition, row_position, 2)

        where_value = QLineEdit()
        layout.addWidget(where_value, row_position, 3)

        delete_button = QPushButton('x')
        delete_button.setMaximumWidth(30)
        delete_button.clicked.connect(lambda: self.delete_where_row(layout, row_position))
        layout.addWidget(delete_button, row_position, 4)
    def add_action_row(self, layout):
        # Get the current number of rows in the layout
        row_position = layout.rowCount()

        action_operation = QComboBox()
        action_operation.addItems(["add", "subtract", "multiply by", "divide by", "set equal to"])
        layout.addWidget(action_operation, row_position, 1)

        action_value = QLineEdit()
        layout.addWidget(action_value, row_position, 2)

        action_filetype_label = QLabel("FOR RASTER:")
        layout.addWidget(action_filetype_label, row_position, 3)

        action_field = QComboBox()
        action_field.addItems(["land cover", "tree cover", "impervious cover"])
        layout.addWidget(action_field, row_position, 4)

        delete_button = QPushButton('x')
        delete_button.setMaximumWidth(30)
        delete_button.clicked.connect(lambda: self.delete_action_row(layout, row_position))
        layout.addWidget(delete_button, row_position, 5)
    #Deletes additional 'Where' and 'Action' fields for each scenario
    def delete_where_row(self, layout, row_position):
        for col in range(layout.columnCount()):
            item = layout.itemAtPosition(row_position, col)
            if item is not None:
                widget = item.widget()
                if widget is not None:
                    widget.deleteLater()
    def delete_action_row(self, layout, row_position):
        for col in range(layout.columnCount()):
            item = layout.itemAtPosition(row_position, col)
            if item is not None:
                widget = item.widget()
                if widget is not None:
                    widget.deleteLater()
    # Scenario - Logic for 'Process Data' JSON output
    def collect_where_data(self, layout):
        where_data = []
        # Get the number of rows in the layout
        rows = layout.rowCount()
        for row in range(rows):
            # Check if operator exists
            if row > 0 and layout.itemAtPosition(row, 0) is not None:
                operator = layout.itemAtPosition(row, 0).widget().currentText()
            else:
                operator = ""

            # Check if where_field exists
            where_field_widget = layout.itemAtPosition(row, 1)
            if where_field_widget is not None:
                where_field_widget = where_field_widget.widget()
                if isinstance(where_field_widget, QComboBox):
                    where_field = where_field_widget.currentText()
                elif isinstance(where_field_widget, QLabel):
                    where_field = where_field_widget.text()
                else:
                    where_field = ""
            else:
                where_field = ""

            # Check if where_condition exists
            where_condition_widget = layout.itemAtPosition(row, 2)
            if where_condition_widget is not None:
                where_condition = where_condition_widget.widget().currentText()
            else:
                where_condition = ""

            # Check if where_value exists
            where_value_widget = layout.itemAtPosition(row, 3)
            if where_value_widget is not None:
                where_value = where_value_widget.widget().text()
            else:
                where_value = ""
            # Only add the row if the value is not empty
            if where_value:
                where_data.append({
                    "operator": operator,
                    "field": where_field,
                    "condition": where_condition,
                    "value": where_value
                })
        return where_data
    def collect_action_data(self, layout):
        action_data = []
        rows = layout.rowCount()
        for row in range(rows):
            # Retrieve the widgets for each field in the layout
            action_operation = layout.itemAtPosition(row, 1).widget().currentText()
            action_value = layout.itemAtPosition(row, 2).widget().text()
            action_field = layout.itemAtPosition(row, 4).widget().currentText()
            # Skip the row if the value is empty
            if action_value:
                action_data.append({
                    "operation": action_operation,
                    "value": action_value,
                    "field": action_field
                })
        return action_data
    #~~~~~~~~~~~~~
    # ADVANCED CONFIG - XML Editor
    #~~~~~~~~~~~~~
    #Creates each parent section for the XML Edtior, sorts into 'recommended' and 'all' variables
    def create_data_folder_with_recommendations(self, root, path, title):
        main_group_box = QGroupBox(title)
        main_layout = QVBoxLayout()
        main_group_box.setLayout(main_layout)

        main_group_box.setStyleSheet("""
            QGroupBox {
                font-size: 10pt;
                font-weight: bold;
            }
        """)

        flagged_variables = ['LeafOnDate_JDay', 'LeafOffDate_JDay', 'LeafTransition_Day',
                             'TreeCanopyCover_overImpervious_frac', 'LeafAreaIndex_Tree_m2pm2', 'LeafAreaIndex_SVeg_m2pm2', 'TreeCover_Evergreen_frac'
                             'Soil_Ksat_mph', 'Soil_WettingFront_Suction_m', 'Soil_WiltingPoint_m3pm3',
                             'Soil_FieldCapacity_m3pm3', 'Soil_SaturationPoint_m3pm3', 'Soil_Porosity_m3pm3',
                             'Soil_MoistureInitial_m3pm3', 'Parameter_m_KsatExpDecay', 'Discharge_Subsurface_Initial_mph',
                             'VadoseZone_Transmissivity_Max_m2ph',
                             'Height_Avg_TreeCover_m', 'Height_Avg_SVegCover_m',
                             'Height_Sensor_Wind_m', 'SVeg_in_Gap_DevelopedArea_frac', 'Resistance_Tree_Min_spm',
                             'Resistance_SVeg_Min_spm']  # Example flagged variables
        calculated_variables = ['DirectlyConnectedImperviousArea_frac', 'Area_m2', 'LeafOnDate_JDay', 'LeafOffDate_JDay']
        calibrated_variables = ['Discharge_Subsurface_Initial_mph', 'VadoseZone_Transmissivity_Max_m2ph', 'Parameter_m_KsatExpDecay', 'TreeCanopyCover_overImpervious_frac']
        # Dictionary to hold the QLineEdit references
        self.line_edits = {}
        recommended_layout = QGridLayout()
        normal_layout = QGridLayout()
        recommended_added = False # Flag to check if any recommended variables are added
        # Find the parent element using XPath
        element = root.xpath(path)
        if not element:
            return main_group_box
        element = element[0]  # Extract the first matching element

        for i, child in enumerate(element):
            # Skip comments
            if isinstance(child, etree._Comment):
                continue

            # Determine the label text
            if child.tag in calculated_variables:
                label_text = child.tag + '*'
            elif child.tag in calibrated_variables:
                label_text = child.tag + '**'
            else:
                label_text = child.tag

            # Ensure label_text is a string
            label_text = str(label_text)

            # Create QLabel
            label = QLabel(label_text)

            # Add tooltip if a comment follows this element
            next_sibling = child.getnext()
            if next_sibling is not None and isinstance(next_sibling, etree._Comment):
                comment_text = next_sibling.text.strip()
                label.setToolTip(comment_text)

            # Create QLineEdit for the value
            line_edit = QLineEdit(child.text.strip() if child.text else "")
            self.line_edits[child.tag] = line_edit

            # Add to the appropriate layout
            if child.tag in flagged_variables:
                recommended_layout.addWidget(label, i, 0)
                recommended_layout.addWidget(line_edit, i, 1)
                recommended_added = True
            else:
                normal_layout.addWidget(label, i, 0)
                normal_layout.addWidget(line_edit, i, 1)

        # Add layouts to the group box
        if recommended_added:
            recommended_group_box = QGroupBox("Recommended Variables")
            recommended_group_box.setLayout(recommended_layout)
            main_layout.addWidget(recommended_group_box)

        normal_group_box = QGroupBox("All Variables")
        normal_group_box.setLayout(normal_layout)
        main_layout.addWidget(normal_group_box)

        return main_group_box
    def create_temp_calc_with_recommendations(self, root, path, title):
        main_group_box = QGroupBox(title)
        main_layout = QVBoxLayout()
        main_group_box.setLayout(main_layout)

        main_group_box.setStyleSheet("""
            QGroupBox {
                font-size: 10pt;
                font-weight: bold;
            }
        """)

        # Dictionary to hold the QLineEdit references
        self.line_edits_temp = {}
        recommended_layout = QGridLayout()
        normal_layout = QGridLayout()
        recommended_added = False # Flag to check if any recommended variables are added
        # Find the parent element using XPath
        element = root.xpath(path)
        if not element:
            return main_group_box
        element = element[0]  # Extract the first matching element

        for i, child in enumerate(element):
            # Skip comments
            if isinstance(child, etree._Comment):
                continue

            # Ensure label_text is a string
            label_text = str(child.tag)

            # Create QLabel
            label = QLabel(label_text)

            # Add tooltip if a comment follows this element
            next_sibling = child.getnext()
            if next_sibling is not None and isinstance(next_sibling, etree._Comment):
                comment_text = next_sibling.text.strip()
                label.setToolTip(comment_text)

            # Create QLineEdit for the value
            line_edit = QLineEdit(child.text.strip() if child.text else "")
            self.line_edits_temp[child.tag] = line_edit

            normal_layout.addWidget(label, i, 0)
            normal_layout.addWidget(line_edit, i, 1)

        normal_group_box = QGroupBox("All Variables")
        normal_group_box.setLayout(normal_layout)
        main_layout.addWidget(normal_group_box)

        return main_group_box
    def create_temp_execution_with_recommendations(self, root, path, title):
        main_group_box = QGroupBox(title)
        #Make group box title larger
        main_group_box.setStyleSheet("""
            QGroupBox {
                font-size: 10pt;
                font-weight: bold;
            }
        """)

        main_layout = QVBoxLayout()
        main_group_box.setLayout(main_layout)
        flagged_variables = ['RefWeatherLocationNLCD_Class', 'RefWeatherLocationTreeCover_percent',
                             'RefWeatherLocationImperviousCover_percent', 'RefWeatherLocationElevation_m']  # Example flagged variables
        calculated_variables = ['TimeBasedOutputPeriod', 'PrintBlockGroupDailyAndHourly', 'RefWeatherLocationElevation_m']
        # Dictionary to hold the QLineEdit references
        self.line_edits_exec = {}
        recommended_layout = QGridLayout()
        normal_layout = QGridLayout()
        recommended_added = False  # Flag to check if any recommended variables are added
        # Find the parent element using XPath
        element = root.xpath(path)
        if not element:
            return main_group_box
        element = element[0]  # Extract the first matching element

        for i, child in enumerate(element):
            # Skip comments
            if isinstance(child, etree._Comment):
                continue

            # Determine the label text
            if child.tag in calculated_variables:
                label_text = child.tag + '*'
            else:
                label_text = child.tag

            # Ensure label_text is a string
            label_text = str(label_text)

            # Create QLabel
            label = QLabel(label_text)

            # Add tooltip if a comment follows this element
            next_sibling = child.getnext()
            if next_sibling is not None and isinstance(next_sibling, etree._Comment):
                comment_text = next_sibling.text.strip()
                label.setToolTip(comment_text)

            # Create QLineEdit for the value
            line_edit = QLineEdit(child.text.strip() if child.text else "")
            self.line_edits_exec[child.tag] = line_edit

            if child.tag in flagged_variables:
                recommended_layout.addWidget(label, i, 0)
                recommended_layout.addWidget(line_edit, i, 1)
                recommended_added = True
            else:
                normal_layout.addWidget(label, i, 0)
                normal_layout.addWidget(line_edit, i, 1)

        if recommended_added:
            recommended_group_box = QGroupBox("Recommended Variables")
            recommended_group_box.setLayout(recommended_layout)
            main_layout.addWidget(recommended_group_box)

        normal_group_box = QGroupBox("All Variables")
        normal_group_box.setLayout(normal_layout)
        main_layout.addWidget(normal_group_box)

        return main_group_box
    #Create scrollable group box for the XML Editor - passed to each parent section
    def create_scrollable_group(self, group_box):
        scroll_area = QScrollArea()
        scroll_area.setWidgetResizable(True)
        scroll_area.setWidget(group_box)
        scroll_area.setMaximumHeight(300)  # Set maximum height for the scroll area
        return scroll_area
    def update_soil_properties(self):
        selected_soil = self.soil_type_dropdown.currentText()
        if selected_soil in self.soil_properties:
            properties = self.soil_properties[selected_soil]
            self.line_edits['Soil_Ksat_mph'].setText(str(properties['Soil_Ksat_mph']))
            self.line_edits['Soil_WettingFront_Suction_m'].setText(str(properties['Soil_WettingFront_Suction_m']))
            self.line_edits['Soil_Porosity_m3pm3'].setText(str(properties['Soil_Porosity_m3pm3']))
            self.line_edits['Soil_FieldCapacity_m3pm3'].setText(str(properties['Soil_FieldCapacity_m3pm3']))
            self.line_edits['Soil_WiltingPoint_m3pm3'].setText(str(properties['Soil_WiltingPoint_m3pm3']))

    # Advanced Config - Logic for 'Process Data' JSON output
    def extract_data(self):
        updated_data = {}
        for tag, line_edit in self.line_edits.items():
            updated_data[tag] = line_edit.text()
        return updated_data
    def extract_data_temp(self):
        updated_data = {}
        for tag, line_edit in self.line_edits_temp.items():
            updated_data[tag] = line_edit.text()
        return updated_data
    def extract_data_exec(self):
        updated_data = {}
        for tag, line_edit in self.line_edits_exec.items():
            updated_data[tag] = line_edit.text()
        return updated_data
    #~~~~~~~~~~~~
    # BROWSE DIALOGS
    #~~~~~~~~~~~~
    #Dialog that appears when 'Browse' button is clicked
    def open_folder_dialog(self, line_edit):
        folder_path = QFileDialog.getExistingDirectory(self, "Select Folder")
        if folder_path:
            line_edit.setText(folder_path)
    def open_shp_file_dialog(self, line_edit):
        file_path, _ = QFileDialog.getOpenFileName(self, "Select File", "", "Shapefile (*.shp);;All Files (*)")
        if file_path:
            line_edit.setText(file_path)
    def open_tif_file_dialog(self, line_edit):
        file_path, _ = QFileDialog.getOpenFileName(self, "Select File", "", "TIFF (*.tif);;All Files (*)")
        if file_path:
            line_edit.setText(file_path)
    def open_csv_file_dialog(self, line_edit):
        file_path, _ = QFileDialog.getOpenFileName(self, "Select File", "", "CSV File (*.csv);;All Files (*)")
        if file_path:
            line_edit.setText(file_path)
    def open_ah_file_dialog(self, line_edit):
        #Open .NC or .TIF files
        file_path, _ = QFileDialog.getOpenFileName(self, "Select File", "", "NetCDF (*.nc);;TIFF (*.tif);;All Files (*)")
        if file_path:
            line_edit.setText(file_path)
    #~~~~~~~~~~~~
    # HELP WINDOW TOGGLES
    #~~~~~~~~~~~~
    def toggle_help_window(self):
        if not self.helpWindow or not self.helpWindow.isVisible():
            self.helpWindow = HelpWindow()
            self.helpWindow.setGeometry(600,100,400,650)
            self.helpWindow.show()
        else:
            self.helpWindow.close()
    def toggle_help_window_adv(self):
        if not self.helpWindow_adv or not self.helpWindow_adv.isVisible():
            self.helpWindow_adv = HelpWindow_Adv()
            self.helpWindow_adv.setGeometry(600,100,400,600)
            self.helpWindow_adv.show()
        else:
            self.helpWindow.close()
    def toggle_help_window_scenario(self):
        if not self.helpWindow_sce or not self.helpWindow_sce.isVisible():
            self.helpWindow_sce = HelpWindow_Sce()
            self.helpWindow_sce.setGeometry(600,100,400,650)
            self.helpWindow_sce.show()
        else:
            self.helpWindow.close()
    #~~~~~~~~~~~~
    # THREAD CLOSING
    #~~~~~~~~~~~~
    def closeEvent(self, event):
        # Ensure all threads are properly terminated
        for thread in self.threads:
            thread.quit()
            thread.wait()
        event.accept()
    #~~~~~~~~~~~~
    # P_ITCA_01
    #~~~~~~~~~~~~
    #Function that runs when 'Process Data' button is clicked on any tab
    def process_data(self):
        #Move to Console tab to display output from script
        self.tab_widget.setTabEnabled(3, True)
        self.tab_widget.setCurrentIndex(3)
        #Clear console output
        self.text_edit.clear()
        #MAPPING GUI TO JSON
        #Proj Details
        proj_name = self.project_name_input.text()
        out_res = self.output_resolution_input.text()
        #if out_res is multiple, make list
        if ',' in out_res:
            output_resolution_list = out_res.split(',')
        else:
            output_resolution_list = [out_res]
        col_row_max = self.col_row_max_input.text()
        sim_start = self.simulation_start_input.text()
        sim_end = self.simulation_end_input.text()
        #Check sim start and end format are both either 8 or 10 digits long
        if len(sim_start) != len(sim_end):
            print("Simulation start and end dates must be the same length")
            return
        if len(sim_start) and len(sim_end) not in [8, 10]:
            print("Simulation start and end dates must be either 8 or 10 digits long")
            return
        flag_OutputEPSG = self.output_epsg_input.text()
        #Weather
        if self.multiple_weather_checkbox.isChecked():
            path_weather = ''
            filepath_multistation = self.multi_station_input.text()
            flag_MultipleStation = True
            if self.hottest_day_checkbox.isChecked():
                flag_HottestDaySimulated = True
            else:
                flag_HottestDaySimulated = False
        else:
            path_weather = self.weather_path_input.text()
            flag_MultipleStation = False
            flag_HottestDaySimulated = False
            filepath_multistation = ''
        #Paths
        path_trunk = self.trunk_path_input.text()
        path_project = self.project_path_input.text()
        filepath_aoi = self.aoi_path_input.text()
        filepath_dem = self.dem_path_input.text()
        dem_units = self.dem_units_dropdown.currentText()
        if dem_units == "cm":
            flag_DEM_ratio = 100.0
        else:
            flag_DEM_ratio = 1.0
        path_nlcd = self.nlcd_path_input.text()
        if self.nlcd_auto_search.isChecked():
            flag_NLCD_AutoSearch = True
            file_lc = ""
            file_ic = ""
            file_tc = ""
            flag_NLCD_LC_year = self.nlcd_auto_search_year.value()
        else:
            flag_NLCD_AutoSearch = False
            file_lc = self.nlcd_land_input.text()
            file_ic = self.nlcd_imperv_input.text()
            file_tc = self.nlcd_tree_input.text()
            flag_NLCD_LC_year = ""
        if self.ah_path_input.text() == "":
            flag_AnthropogenicHeat = False
            filepath_ah4guc = ""
        else:
            flag_AnthropogenicHeat = True
            filepath_ah4guc = self.ah_path_input.text()
        if self.census_path_input.text() == "":
            flag_Census = False
            filepath_census = ""
            census_fids = ""
        else:
            flag_Census = True
            filepath_census = self.census_path_input.text()
            census_fids = self.census_list_input.text()

        #Advanced Config Flags
        flag_XML_Import = self.xml_import_checkbox.isChecked()
        flag_NumberofBatches = self.batch_number_spinbox.value()
        flag_DEM_average = self.dem_avg_checkbox.isChecked()
        soil_type = self.soil_type_dropdown.currentText()
        flag_Hydro = self.hydro_prep_checkbox.isChecked()
        if flag_Hydro:
            # Get the project path and replace the last folder with 'Hydro'
            path_project = self.project_path_input.text()
            base_path = Path(path_project).parent
            hydro_path = base_path / 'Hydro'
        else:
            hydro_path = ''
        flag_Calibration = self.calibration_checkbox.isChecked()
        if flag_Calibration:
            flag_Vadose_auto = self.vadose_zone_checkbox.isChecked()
            if not flag_Vadose_auto:
                first_vadose = self.start_vadose_zone_input.text()
                last_vadose = self.end_vadose_zone_input.text()
            else:
                first_vadose = ''
                last_vadose = ''
            calibration_start = self.calibration_start_input.text()
            calibration_end = self.calibration_end_input.text()
        else:
            flag_Vadose_auto = ''
            first_vadose = ''
            last_vadose = ''
            calibration_start = ''
            calibration_end = ''

        #Open HydroPlusConfigXML and read in variables for parents
        parser = etree.XMLParser(remove_blank_text=True)
        tree = etree.parse(current_dir / 'resources' / 'HydroPlusConfig_CoolAir_Template.xml', parser)
        root = tree.getroot()
        updated_data = self.extract_data()
        updated_data_temp = self.extract_data_temp()
        updated_data_exec = self.extract_data_exec()
        paths = ['DataOrganizer/DataDrawer/DataFolder', 'TemperatureCalculationParams', 'TemperatureExecutionParams']
        xml_import = {}
        for path in paths:
            parent = root.find(path)
            xml_import[parent.tag] = {}
            #If the child text (from HydroPlusConfig.xml) is different from the user-end values, store in dictionary
            for child in parent:
                if child.tag in updated_data and child.text != updated_data[child.tag]:
                    xml_import[parent.tag][child.tag] = updated_data[child.tag]
                elif child.tag in updated_data_temp and child.text != updated_data_temp[child.tag]:
                    xml_import[parent.tag][child.tag] = updated_data_temp[child.tag]
                elif child.tag in updated_data_exec and child.text != updated_data_exec[child.tag]:
                    xml_import[parent.tag][child.tag] = updated_data_exec[child.tag]
        #Scenarios
        all_scenarios_data = []
        for i, scenario_box in enumerate(self.scenario_boxes):
            #print(f"Extracting data from Scenario {i + 1}")
            scenario_data = {
                "name": "",
                "where": [],
                "action": []
            }

            # Iterate over all child widgets of the scenario_box
            for group_box in scenario_box.findChildren(QGroupBox):
                #print(f"Group box title: {group_box.title()}"
                      #f"Scenario box = {scenario_box}")
                if group_box.title() == "WHERE:":
                    where_layout = group_box.layout()
                    where_data = self.collect_where_data(where_layout)
                    scenario_data["where"].extend(where_data)
                elif group_box.title() == "ACTION:":
                    action_layout = group_box.layout()
                    action_data = self.collect_action_data(action_layout)
                    scenario_data["action"].extend(action_data)
                scenario_name = scenario_box.findChild(QLineEdit).text()
                scenario_data["name"] = scenario_name


            if scenario_data["where"] or scenario_data["action"]:
                #print(f"Scenario {i + 1} data: {scenario_data}")
                all_scenarios_data.append(scenario_data)


        #Package all variables into a dictionary
        config_data = {
            'proj_name': proj_name,
            'output_resolution_list': output_resolution_list,
            'col_max': col_row_max,
            'row_max': col_row_max,
            'simulation_start': sim_start,
            'simulation_end': sim_end,
            'flag_OutputEPSG': flag_OutputEPSG,
            'flag_MultipleStation': flag_MultipleStation,
            'flag_HottestDaySimulated': flag_HottestDaySimulated,
            'path_weather': path_weather,
            'filepath_multistation': filepath_multistation,
            'path_trunk': path_trunk,
            'path_project': path_project,
            'filepath_aoi': filepath_aoi,
            'filepath_dem': filepath_dem,
            'dem_units': dem_units,
            'flag_DEM_ratio': flag_DEM_ratio,
            'path_nlcd': path_nlcd,
            'flag_NLCDAutoSearch': flag_NLCD_AutoSearch,
            'flag_NLCD_LC_year': flag_NLCD_LC_year,
            'file_lc': file_lc,
            'file_ic': file_ic,
            'file_tc': file_tc,
            'flag_AnthropogenicHeat': flag_AnthropogenicHeat,
            'filepath_ah4guc': filepath_ah4guc,
            'flag_Census': flag_Census,
            'filepath_census': filepath_census,
            'census_fids': census_fids,
            'flag_XML_Import': flag_XML_Import,
            'flag_NumberofBatches': flag_NumberofBatches,
            'flag_DEM_average': flag_DEM_average,
            'flag_Hydro': flag_Hydro,
            'hydro_path': str(hydro_path),
            'xml_import': xml_import,
            'soil_type': soil_type,
            'flag_Calibration': flag_Calibration,
            'flag_Vadose_auto': flag_Vadose_auto,
            'first_vadose': first_vadose,
            'last_vadose': last_vadose,
            'calibration_start': calibration_start,
            'calibration_end': calibration_end,
            'all_scenarios_data': all_scenarios_data
        }
        #Navigate to json folder and write config_data to json file
        json_folder = current_dir / 'json'
        if not json_folder.exists():
            json_folder.mkdir(parents=True, exist_ok=True)
        filename = json_folder / f'{proj_name}_CoolAirConfig.json'
        filename_fs = str(filename).replace('\\', '/')
        with open(filename, 'w') as f:
            json.dump(config_data, f, indent=4)
        #Run p_itca_01
        self.start_script(proj_name, filename_fs)
    #Run p_itca_01 script
    def start_script(self, proj_name, filename):
        command = [sys.executable, '-u', current_dir / 'scripts' / 'p_itca_01_prep.py', '--proj_name', proj_name,
                   '--config_filename', filename]
        # Create a new worker thread to run the command
        self.thread = WorkerThread(command)
        # Connect the signals to the appropriate slots
        #Updates console with output from p_itca_01
        self.thread.output_signal.connect(self.update_console)
        #Updates batch_files list with output from p_itca_01 that ends in .bat
        self.thread.batch_files_signal.connect(self.handle_batch_toggle)
        #Updates output_folders list with output from p_itca_01 that starts with "output folder="
        self.thread.output_folder_signal.connect(self.handle_output_folder)
        self.terminate_button.setEnabled(True)

        self.thread.start()
    #Sends output text from p_itca_01 to console
    def update_console(self, message):
        # Append the output message to the text edit
        self.text_edit.append(message)
    # If output folder is received from p_itca_01, append folder to output_folders for use in auto-run mosaic function
    def handle_output_folder(self, folder):
        self.output_folders.append(folder)

    def terminate_script(self):
        if self.thread is not None:
            self.thread.terminate()
            self.thread.wait()  # Wait for the thread to finish
            self.thread = None
            self.update_console("Script terminated.")
            self.terminate_button.setEnabled(False)
    #~~~~~~~~~~~~
    # COOL AIR HANDLER
    #~~~~~~~~~~~~
    #Runs batches, executed by 'Run Cool Air' button
    def run_batches(self):
        #Create a new thread for each batch file
        self.batch_threads = []
        for batch_file in self.batch_files:
            batch_command = f'cmd /c start {batch_file}'
            thread = BatchThread(batch_command)
            thread.finished_signal.connect(self.handle_batch_finish)
            thread.start()
            self.batch_threads.append(thread)
    #Updates console with complete batch message
    def handle_batch_finish(self, message):
        print(message)
    #~~~~~~~~~~~~
    # P_ITCA_03
    #~~~~~~~~~~~~
    #Updates console with output from p_itca_03
    def update_console_mosaic(self, message):
        # Append the output message to the text edit
        self.text_edit_mosaic.append(message)
    #If output_folder(s) is received from p_itca_01, auto-run mosaics is enabled based on input YYYYMMDDHH + project name from user
    def auto_run_mosaic(self):
        #User input for temperture units
        celsius_check = self.celsius_checkbox.isChecked()
        fahrenheit_check = self.fahrenheit_checkbox.isChecked()
        kelvin_check = self.kelvin_checkbox.isChecked()
        startday = self.simulation_start_input.text()
        endday = self.simulation_end_input.text()
        #Add on 00 or 23 to start and end day if user only inputs YYYYMMDD
        if len(startday) == 8:
            startday += '00'
        if len(endday) == 8:
            endday += '23'

        #Pack temperature units into dictionary
        temp_units = {
            'celsius': celsius_check,
            'fahrenheit': fahrenheit_check,
            'kelvin': kelvin_check
        }

        #Clear console
        self.text_edit_mosaic.clear()
        self.text_edit_mosaic.setText("================================\n"
                                      "Mosaic Processor Console\n"
                                      "================================\n"
                                      "Beginning Mosaic Processing...\n")
        #Stores commands to be run
        self.mosaic_commands = []
        if self.output_folders:
            for output_folder in self.output_folders:
                command = [sys.executable, '-u', current_dir / 'scripts' / 'p_itca_03_mosaic.py',
                           '--output_folder', output_folder, '--output_epsg', self.output_epsg_input.text(),
                           '--startday', startday, '--endday', endday,
                           '--temp_units', json.dumps(temp_units)]
                self.mosaic_commands.append(command)
            self.run_next_mosaic()
        else:
            print("No output folder found. Please manually run the mosaic script.")
    #Run next mosaic command in the list generated by auto_run_mosaic
    def run_next_mosaic(self):
        if self.mosaic_commands:
            command = self.mosaic_commands.pop(0)
            mosaic_thread = MosaicWorkerThread(command)
            mosaic_thread.output_signal.connect(self.update_console_mosaic)
            mosaic_thread.finished.connect(self.on_mosaic_finished)
            self.threads.append(mosaic_thread)
            mosaic_thread.start()
        else:
            print("All mosaic tasks completed.")
    #Remove finished thread from list and run next mosaic
    def on_mosaic_finished(self):
        self.threads.pop(0)  # Remove the finished thread from the list
        self.run_next_mosaic()  # Run the next mosaic
    #Manually run mosaic script with user input
    def manual_run_mosaic(self):
        #Get user inputs
        output_folder = self.mosaic_folder_output.text()
        output_epsg = self.mosaic_epsg_input.text()
        startday = self.mosaic_start_input.text()
        endday = self.mosaic_end_input.text()
        #Add on 00 or 23 to start and end day if user only inputs YYYYMMDD
        if len(startday) == 8:
            startday += '00'
        if len(endday) == 8:
            endday += '23'
        #Get user temperature units
        celsius_check = self.celsius_checkbox.isChecked()
        fahrenheit_check = self.fahrenheit_checkbox.isChecked()
        kelvin_check = self.kelvin_checkbox.isChecked()
        #Pack checks into dictionary
        temp_units = {
            'celsius': celsius_check,
            'fahrenheit': fahrenheit_check,
            'kelvin': kelvin_check
        }
        #Clear console
        self.text_edit_mosaic.clear()
        self.text_edit_mosaic.setText("================================\n"
                                      "Mosaic Processor Console\n"
                                      "================================\n"
                                      "Beginning Mosaic Processing...\n")
        if output_folder:
            command = [sys.executable, '-u', current_dir / 'scripts' / 'p_itca_03_mosaic.py', '--output_folder',
                       output_folder, '--output_epsg', output_epsg,
                       '--startday', startday, '--endday', endday,
                       '--temp_units', json.dumps(temp_units)]
        self.mosaic_thread = MosaicWorkerThread(command)
        self.mosaic_thread.output_signal.connect(self.update_console_mosaic)
        self.mosaic_thread.start()
    #~~~~~~~~~~~~
    # GUI
    #~~~~~~~~~~~~
    def __init__(self):
        super().__init__()
        self.setWindowTitle("i-Tree Cool Air")  # Set the window title
        self.setGeometry(100, 100, 500, 600)  # Set position and dimensions (x, y, width, height)
        icon_path = current_dir / 'resources' / 'iTree_transparent.ico'
        self.setWindowIcon(QIcon(str(icon_path)))
        #List of XML Variables that are commonly adjusted
        self.tab_widget = QTabWidget(self)  # Create the tab widget
        self.setCentralWidget(self.tab_widget)  # Set the tab widget as the central widget
        self.create_tabs()  # Method to create tabs
        self.output_folders = []
        self.batch_files = []
        self.threads = []
        self.helpWindow = HelpWindow()
        self.helpWindow_adv = HelpWindow_Adv()
        self.helpWindow_sce = HelpWindow_Sce()

        self.soil_properties = {
            "Sand": {"Soil_Ksat_mph": 0.12, "Soil_WettingFront_Suction_m": 0.1, "Soil_Porosity_m3pm3": 0.437,
                     "Soil_FieldCapacity_m3pm3": 0.062, "Soil_WiltingPoint_m3pm3": 0.024},
            "Loamy sand": {"Soil_Ksat_mph": 0.03, "Soil_WettingFront_Suction_m": 0.1555, "Soil_Porosity_m3pm3": 0.437,
                           "Soil_FieldCapacity_m3pm3": 0.105, "Soil_WiltingPoint_m3pm3": 0.047},
            "Sandy loam": {"Soil_Ksat_mph": 0.011, "Soil_WettingFront_Suction_m": 0.18, "Soil_Porosity_m3pm3": 0.453,
                           "Soil_FieldCapacity_m3pm3": 0.19, "Soil_WiltingPoint_m3pm3": 0.085},
            "Loam": {"Soil_Ksat_mph": 0.003, "Soil_WettingFront_Suction_m": 0.2195, "Soil_Porosity_m3pm3": 0.463,
                     "Soil_FieldCapacity_m3pm3": 0.232, "Soil_WiltingPoint_m3pm3": 0.116},
            "Silt loam": {"Soil_Ksat_mph": 0.007, "Soil_WettingFront_Suction_m": 0.17, "Soil_Porosity_m3pm3": 0.501,
                          "Soil_FieldCapacity_m3pm3": 0.284, "Soil_WiltingPoint_m3pm3": 0.135},
            "Sandy clay loam": {"Soil_Ksat_mph": 0.002, "Soil_WettingFront_Suction_m": 0.22,
                                "Soil_Porosity_m3pm3": 0.398, "Soil_FieldCapacity_m3pm3": 0.244,
                                "Soil_WiltingPoint_m3pm3": 0.136},
            "Clay loam": {"Soil_Ksat_mph": 0.001, "Soil_WettingFront_Suction_m": 0.21, "Soil_Porosity_m3pm3": 0.464,
                          "Soil_FieldCapacity_m3pm3": 0.31, "Soil_WiltingPoint_m3pm3": 0.187},
            "Silty clay loam": {"Soil_Ksat_mph": 0.001, "Soil_WettingFront_Suction_m": 0.27,
                                "Soil_Porosity_m3pm3": 0.471, "Soil_FieldCapacity_m3pm3": 0.342,
                                "Soil_WiltingPoint_m3pm3": 0.21},
            "Sandy clay": {"Soil_Ksat_mph": 0.001, "Soil_WettingFront_Suction_m": 0.24, "Soil_Porosity_m3pm3": 0.43,
                           "Soil_FieldCapacity_m3pm3": 0.321, "Soil_WiltingPoint_m3pm3": 0.221},
            "Silty clay": {"Soil_Ksat_mph": 0.001, "Soil_WettingFront_Suction_m": 0.29, "Soil_Porosity_m3pm3": 0.479,
                           "Soil_FieldCapacity_m3pm3": 0.371, "Soil_WiltingPoint_m3pm3": 0.251},
            "Clay": {"Soil_Ksat_mph": 0.0003, "Soil_WettingFront_Suction_m": 0.66, "Soil_Porosity_m3pm3": 0.475,
                     "Soil_FieldCapacity_m3pm3": 0.378, "Soil_WiltingPoint_m3pm3": 0.265}
        }

    def create_tabs(self):
        #===============================
        # TAB 1: Project Set-Up
        #===============================
        self.tab1 = QWidget()
        tab1_layout = QVBoxLayout(self.tab1)

        #Title
        title_label = QLabel("i-Tree Cool Air: Project Set-Up")
        title_label.setStyleSheet("""
            QLabel {
                font-size: 16pt;
            }
        """)
        title_group_layout = QGridLayout()
        title_group_layout.addWidget(title_label, 0, 0)

        #JSON Import
        json_import = QPushButton("Import JSON")
        json_import.clicked.connect(self.import_json)
        title_group_layout.addWidget(json_import, 0, 2)
        json_import.setMaximumWidth(100)

        #Help Button
        help_button = QPushButton("Help")
        help_button.clicked.connect(self.toggle_help_window)
        title_group_layout.addWidget(help_button, 0, 3)
        help_button.setMaximumWidth(50)

        #~~~~~~~~~~~~
        # Project Details Section
        # ~~~~~~~~~~~~
        #Title
        project_group_box = QGroupBox("Project Details")
        project_group_layout = QGridLayout()
        project_group_box.setLayout(project_group_layout)
        project_group_box.setStyleSheet("""
            QGroupBox {
                font-size: 12pt; 
                font-weight: bold;
            }
        """)
        # Project Name - LineEdit
        project_name_label = QLabel("Project Name:")
        self.project_name_input = QLineEdit("36067_gui_test")
        project_group_layout.addWidget(project_name_label, 0, 0)  # Position at row 0, column 0
        project_group_layout.addWidget(self.project_name_input, 0, 1)  # Position at row 0, column 1

        # Simulation Start - LineEdit / Integer
        simulation_start_label = QLabel("Simulation Start:")
        simulation_start_label.setToolTip("Can be format YYYYMMDD or YYYYMMDDHH\n"
                                          "e.g. 20210101 or 2021010100")
        self.simulation_start_input = QLineEdit("20210701")
        project_group_layout.addWidget(simulation_start_label, 1, 0)
        project_group_layout.addWidget(self.simulation_start_input, 1, 1)

        # Simulation End - LineEdit / Integer
        simulation_end_label = QLabel("Simulation End:")
        simulation_end_label.setToolTip("Can be format YYYYMMDD or YYYYMMDDHH\n"
                                          "e.g. 20210101 or 2021010100")
        self.simulation_end_input = QLineEdit("20210702")
        project_group_layout.addWidget(simulation_end_label, 2, 0)
        project_group_layout.addWidget(self.simulation_end_input, 2, 1)

        # Col/Row Max - LineEdit / Integer / with Validator
        col_row_max_label = QLabel("Col / Row Max:")
        col_row_max_label.setToolTip("Maximum number of columns and rows for the simulation grid, up to 300.")
        self.col_row_max_input = QLineEdit("150")  # Default integer value of 150
        self.col_row_max_input.setValidator(QIntValidator(0, 300))  # Set validator to only allow integers between 0 and 300
        project_group_layout.addWidget(col_row_max_label, 0, 2)
        project_group_layout.addWidget(self.col_row_max_input, 0, 3)

        # Output Resolution - LineEdit / Integer
        output_resolution_label = QLabel("Output Resolution(s):")
        output_resolution_label.setToolTip("Comma-separated list of output resolutions, in meters, for the simulation grid.")
        self.output_resolution_input = QLineEdit("30")  # Default fill-in of 30
        project_group_layout.addWidget(output_resolution_label, 1, 2)
        project_group_layout.addWidget(self.output_resolution_input, 1, 3)

        # Output EPSG - LineEdit / Integer
        output_epsg_label = QLabel("Output EPSG:")
        output_epsg_label.setToolTip("EPSG code for the output projection\n"
                                     "5070 = CONUS Albers Equal Area\n"
                                     "326## = WGS 84/ UTM Zone ##N\n"
                                     "9820 = Lambert Azimuthal Equal Area")
        self.output_epsg_input = QLineEdit("5070")  # Default fill-in of 5070
        project_group_layout.addWidget(output_epsg_label, 2, 2)
        project_group_layout.addWidget(self.output_epsg_input, 2, 3)

        # ~~~~~~~~~~~~
        # Weather Section
        # ~~~~~~~~~~~~
        #Title
        flags_group_box = QGroupBox("Weather")
        flags_group_layout = QGridLayout()
        flags_group_box.setLayout(flags_group_layout)
        flags_group_box.setStyleSheet("""
            QGroupBox {
                font-size: 12pt; 
                font-weight: bold;
            }
        """)
        # Multiple Weather Stations - Checkbox, default unchecked,
        #
        # Enables Multi-Station CSV & Hottest Day Simulated, disables Weather Path
        self.multiple_weather_checkbox = QCheckBox("Multiple Weather Stations")
        self.multiple_weather_checkbox.setToolTip("Use multiple weather station algorithm & CSV")
        flags_group_layout.addWidget(self.multiple_weather_checkbox, 0, 0, 1, 2)  # Span across two columns

        # Add checkbox for 'Hottest Day Simulated'
        self.hottest_day_checkbox = QCheckBox("Hottest Day Simulated")
        self.hottest_day_checkbox.setToolTip("Simulate the hottest day for each station. This overrides the Simulation Start and End.")
        flags_group_layout.addWidget(self.hottest_day_checkbox, 0, 2, 1, 2)  # Span across two columns
        #Disable the hottest day simulated checkbox
        self.hottest_day_checkbox.setEnabled(False)
        #Move hottest day checkbox to right

        # Paths section
        weather_path_label = QLabel("Weather path:")
        weather_path_label.setToolTip("Path to the folder containing Weather.csv and Radiation.csv files.")
        self.weather_path_input = QLineEdit("C:/iTree/WeatherPrep_SVN/trunk/Outputs/725190-14771/2021/")
        weather_path_button = QPushButton("Browse...")
        weather_path_button.clicked.connect(lambda: self.open_folder_dialog(self.weather_path_input))
        flags_group_layout.addWidget(weather_path_label, 1, 0)
        flags_group_layout.addWidget(self.weather_path_input, 1, 1, 1, 2)
        flags_group_layout.addWidget(weather_path_button, 1, 3)

        #Multi station selection
        multi_station_label = QLabel("Multi-Station CSV:")
        multi_station_label.setToolTip("Path to the CSV file containing multi weather station data.")
        self.multi_station_input = QLineEdit()
        multi_station_button = QPushButton("Browse...")
        multi_station_button.clicked.connect(lambda: self.open_csv_file_dialog(self.multi_station_input))
        flags_group_layout.addWidget(multi_station_label, 2, 0)
        flags_group_layout.addWidget(self.multi_station_input, 2, 1, 1, 2)
        flags_group_layout.addWidget(multi_station_button, 2, 3)
        #Set to disabled by default
        self.multi_station_input.setEnabled(False)

        paths_group_box = QGroupBox("Paths Configuration")
        paths_group_layout = QGridLayout()
        paths_group_box.setLayout(paths_group_layout)

        paths_group_box.setStyleSheet("""
            QGroupBox {
                font-size: 12pt;
                font-weight: bold;
            }
        """)

        # Trunk path with folder selection button
        trunk_path_label = QLabel("Project Set-Up Folder:")
        trunk_path_label.setToolTip("Path to intermediate folder that stores TIF and shapefiles before quad creation.")
        self.trunk_path_input = QLineEdit("P:/iTree/a_prep_DEM_NLCD/z_places/")
        trunk_path_button = QPushButton("Browse...")
        trunk_path_button.clicked.connect(lambda: self.open_folder_dialog(self.trunk_path_input))
        paths_group_layout.addWidget(trunk_path_label, 0, 0)
        paths_group_layout.addWidget(self.trunk_path_input, 0, 1, 1, 2)
        paths_group_layout.addWidget(trunk_path_button, 0, 3)


        # Project path with folder selection button
        project_path_label = QLabel("Project Folder:")
        project_path_label.setToolTip("Path to the project folder where the inputs and outputs will be saved. Last parent folder should be 'CoolAir'.")
        self.project_path_input = QLineEdit("P:/iTree/projects/CoolAir")
        project_path_button = QPushButton("Browse...")
        project_path_button.clicked.connect(lambda: self.open_folder_dialog(self.project_path_input))
        paths_group_layout.addWidget(project_path_label, 1, 0)
        paths_group_layout.addWidget(self.project_path_input, 1, 1, 1, 2)
        paths_group_layout.addWidget(project_path_button, 1, 3)

        # AOI filepath with file selection button
        aoi_path_label = QLabel("AOI Filepath:")
        aoi_path_label.setToolTip("Path to the shapefile of the area of interest (AOI)")
        self.aoi_path_input = QLineEdit('P:/iTree/a_prep_DEM_NLCD/Shapefile_test/36067.shp')
        aoi_path_button = QPushButton("Browse...")
        aoi_path_button.clicked.connect(lambda: self.open_shp_file_dialog(self.aoi_path_input))
        paths_group_layout.addWidget(aoi_path_label, 2, 0)
        paths_group_layout.addWidget(self.aoi_path_input, 2, 1, 1, 2)
        paths_group_layout.addWidget(aoi_path_button, 2, 3)

        # DEM filepath with file selection button and dropdown for units
        dem_path_label = QLabel("DEM Filepath:")
        dem_path_label.setToolTip("Path to the digital elevation model (DEM) file, often in .tif format")
        self.dem_path_input = QLineEdit('P:/iTree/a_prep_DEM_NLCD/DEM/NHDComposites/FIPS36/elev_cm.tif')
        dem_path_button = QPushButton("Browse...")

        self.dem_units_dropdown = QComboBox()
        dem_units_label = QLabel("DEM Units:")
        dem_units_label.setToolTip("Units of the DEM file, either in centimeters (cm) or meters (m)")
        self.dem_units_dropdown.addItems(["cm", "m"])
        self.dem_units_dropdown.setCurrentText("cm")
        dem_path_button.clicked.connect(lambda: self.open_tif_file_dialog(self.dem_path_input))
        paths_group_layout.addWidget(dem_path_label, 3, 0)
        paths_group_layout.addWidget(self.dem_path_input, 3, 1, 1, 2)
        paths_group_layout.addWidget(dem_path_button, 3, 3)
        paths_group_layout.addWidget(dem_units_label, 4, 2)
        paths_group_layout.addWidget(self.dem_units_dropdown, 4, 3)

        # Set alignment of the DEM units label to the right
        dem_units_label.setAlignment(Qt.AlignRight | Qt.AlignVCenter)

        #NLCD Filepath
        nlcd_path_label = QLabel("Land Cover Folder:")
        nlcd_path_label.setToolTip("Path to the folder containing the land, impervious, and tree cover files")
        self.nlcd_path_input = QLineEdit('P:/iTree/a_prep_DEM_NLCD/NLCD_test/')
        nlcd_path_button = QPushButton("Browse...")
        nlcd_path_button.clicked.connect(lambda: self.open_folder_dialog(self.nlcd_path_input))
        paths_group_layout.addWidget(nlcd_path_label, 5, 0)
        paths_group_layout.addWidget(self.nlcd_path_input, 5, 1, 1, 2)
        paths_group_layout.addWidget(nlcd_path_button, 5, 3)
        #Checkbox to auto-search for NLCD
        self.nlcd_auto_search = QCheckBox("Auto-search NLCD")
        self.nlcd_auto_search.setToolTip("Automatically search for NLCD filenames using the prefix NLCD_yyyy_...")
        paths_group_layout.addWidget(self.nlcd_auto_search, 6, 2, 1, 1)
        self.nlcd_auto_search_year = QSpinBox()
        self.nlcd_auto_search_year.setRange(2001, 2021)
        self.nlcd_auto_search_year.setValue(2019)
        self.nlcd_auto_search_year.setEnabled(False)
        paths_group_layout.addWidget(self.nlcd_auto_search_year, 6, 3)


        #Manual filepath names if auto-search is not selected
        nlcd_land_label = QLabel("Land Cover Filename:")
        self.nlcd_land_input = QLineEdit("landcover.tif")
        paths_group_layout.addWidget(nlcd_land_label, 7, 0)
        paths_group_layout.addWidget(self.nlcd_land_input, 7, 1, 1, 2)

        nlcd_imperv_label = QLabel("Impervious Filename:")
        self.nlcd_imperv_input = QLineEdit("impervious.tif")
        paths_group_layout.addWidget(nlcd_imperv_label, 8, 0)
        paths_group_layout.addWidget(self.nlcd_imperv_input, 8, 1, 1, 2)

        nlcd_tree_label = QLabel("Tree Cover Filename:")
        self.nlcd_tree_input = QLineEdit("treecover.tif")
        paths_group_layout.addWidget(nlcd_tree_label, 9, 0)
        paths_group_layout.addWidget(self.nlcd_tree_input, 9, 1, 1, 2)

        #Optional inputs QGrid Box
        optional_input_box = QGroupBox("Optional Inputs")
        optional_input_layout = QGridLayout()
        optional_input_box.setLayout(optional_input_layout)
        optional_input_box.setStyleSheet("""
            QGroupBox {
                font-size: 10pt;
                font-weight: bold;
            }
        """)

        #Optional AH File
        ah_path_label = QLabel("Anthropogenic Heat Filepath:")
        ah_path_label.setToolTip("OPTIONAL: Path to the anthropogenic heat file, .NC (multi-timestep) or .TIF (single-timestep) format")
        self.ah_path_input = QLineEdit()
        ah_path_button = QPushButton("Browse...")
        ah_path_button.clicked.connect(lambda: self.open_ah_file_dialog(self.ah_path_input))
        optional_input_layout.addWidget(ah_path_label, 0, 0)
        optional_input_layout.addWidget(self.ah_path_input, 0, 1, 1, 2)
        optional_input_layout.addWidget(ah_path_button, 0, 3)

        #Optional Census File
        census_path_label = QLabel("Census Filepath:")
        census_path_label.setToolTip("OPTIONAL: Path to a shapefile splitting the AOI into subregions, often used for Census block groups for individual processing")
        self.census_path_input = QLineEdit()
        census_path_button = QPushButton("Browse...")
        census_path_button.clicked.connect(lambda: self.open_shp_file_dialog(self.census_path_input))
        optional_input_layout.addWidget(census_path_label, 1, 0)
        optional_input_layout.addWidget(self.census_path_input, 1, 1, 1, 2)
        optional_input_layout.addWidget(census_path_button, 1, 3)

        #Optional: FID List for Census File
        census_list_label = QLabel("Census FID List:")
        census_list_label.setToolTip("OPTIONAL: List of FIDs to process from the Census file. Default is 'All'")
        self.census_list_input = QLineEdit("All")
        optional_input_layout.addWidget(census_list_label, 2, 0)
        optional_input_layout.addWidget(self.census_list_input, 2, 1, 1, 2)
        #Set to disabled by default
        self.census_list_input.setEnabled(False)

        paths_group_layout.addWidget(optional_input_box, 10, 0, 4, 4)


        #Process Button
        process_box = QGroupBox("")
        process_layout = QGridLayout()
        process_box.setLayout(process_layout)
        self.process_button = QPushButton("Process")
        self.process_button.clicked.connect(self.process_data)
        process_layout.addWidget(self.process_button,0, 1, 1, 2)
        #Add widgets to tab1 layout
        tab1_layout.addLayout(title_group_layout)
        tab1_layout.addWidget(project_group_box)
        tab1_layout.addWidget(flags_group_box)
        tab1_layout.addWidget(paths_group_box)
        tab1_layout.addWidget(process_box)

        #Set Stretch
        tab1_layout.setStretch(0, 1)
        tab1_layout.setStretch(1, 1)
        tab1_layout.setStretch(2, 3)

        # Checkbox logic
        self.multiple_weather_checkbox.toggled.connect(self.handle_multiple_weather_toggle)
        self.nlcd_auto_search.toggled.connect(self.handle_nlcd_autosearch_toggle)
        #Census fill-in logic
        self.census_path_input.textChanged.connect(self.update_census_list_toggle)
        self.census_list_input.textChanged.connect(self.update_census_list)
        # Add tab to the widget
        self.tab_widget.addTab(self.tab1, "Project Set-Up")
        #===============================
        # TAB 2: Advanced Config
        #===============================
        #Advanced Config tab
        self.tab2 = QWidget()  # Create a QWidget for the tab
        tab2_layout = QVBoxLayout(self.tab2)

        # Title and Help Section
        adv_label = QLabel("i-Tree Cool Air: Advanced Configuration")
        adv_label.setStyleSheet("""
                    QLabel {
                        font-size: 16pt;
                    }
                """)
        adv_title_group_layout = QGridLayout()
        adv_title_group_layout.addWidget(adv_label, 0, 0)

        help_button_adv = QPushButton("Help")
        help_button_adv.clicked.connect(self.toggle_help_window_adv)
        adv_title_group_layout.addWidget(help_button_adv, 0, 3)
        # Make button less wide
        help_button_adv.setMaximumWidth(50)
        #Advanced Config Group
        advanced_group_box = QGroupBox("Cool Air Settings")
        advanced_group_layout = QGridLayout()
        advanced_group_box.setLayout(advanced_group_layout)
        advanced_group_box.setStyleSheet("""
            QGroupBox {
                font-size: 12pt;
                font-weight: bold;
            }
        """)
        #Calculate RefWeatherLocationElevation_m checkbox
        self.dem_avg_checkbox = QCheckBox("Calculate RefWeatherLocationElevation_m")
        self.dem_avg_checkbox.setToolTip("Calculate average elevation for the reference weather location\n"
                                         "If unchecked, set RefWeatherLocationElevation_m in XML Configuration")
        advanced_group_layout.addWidget(self.dem_avg_checkbox, 0, 0, 1, 2)
        #Batch Number Spinbox
        self.batch_number_label = QLabel("Parallel Batches")
        self.batch_number_spinbox = QSpinBox()
        self.batch_number_label.setToolTip("Number of batches to run the simulation at once")
        self.batch_number_spinbox.setMinimum(1)
        self.batch_number_spinbox.setMaximum(10)
        advanced_group_layout.addWidget(self.batch_number_label, 0, 2)
        advanced_group_layout.addWidget(self.batch_number_spinbox, 0, 3)
        self.batch_number_label.setAlignment(Qt.AlignRight | Qt.AlignVCenter)
        #Move the spinbox to the left and scale size down
        #self.batch_number_spinbox.setAlignment(Qt.AlignRight | Qt.AlignVCenter)
        self.batch_number_spinbox.setFixedWidth(50)

        # Soil Type Dropdown
        self.soil_type_label = QLabel("Preset Soil Type:")
        self.soil_type_label.setToolTip("[Optional] Select a preset soil type to populate soil properties in DataFolder")
        self.soil_type_dropdown = QComboBox()
        self.soil_type_dropdown.addItems([
            "", "Sand", "Loamy sand", "Sandy loam", "Loam", "Silt loam",
            "Sandy clay loam", "Clay loam", "Silty clay loam",
            "Sandy clay", "Silty clay", "Clay"
        ])
        self.soil_type_dropdown.setToolTip("Select the soil type for the simulation")
        self.soil_type_dropdown.currentIndexChanged.connect(self.update_soil_properties)

        advanced_group_layout.addWidget(self.soil_type_label, 1, 0)
        advanced_group_layout.addWidget(self.soil_type_dropdown, 1, 1)


        # Hydro & Calibration Group
        hydro_group_box = QGroupBox("Hydro Calibration")
        hydro_group_layout = QGridLayout()
        hydro_group_box.setLayout(hydro_group_layout)
        hydro_group_box.setStyleSheet("""
            QGroupBox {
                font-size: 12pt;
                font-weight: bold;
            }
        """)

        # Generate Hydro Inputs and Calibration Checkboxes on the same line
        self.hydro_prep_checkbox = QCheckBox("Generate Hydro Inputs")
        self.hydro_prep_checkbox.setToolTip("Generate Hydro Inputs for the simulation, saved in Project Folder under 'Hydro'")
        self.hydro_prep_checkbox.toggled.connect(self.toggle_hydro_fields)
        hydro_group_layout.addWidget(self.hydro_prep_checkbox, 0, 0)

        self.calibration_checkbox = QCheckBox("Calibration")
        self.calibration_checkbox.setToolTip("Run PEST Calibration for the simulation (recommended if generating Hydro inputs)")
        self.calibration_checkbox.setEnabled(False)
        self.calibration_checkbox.toggled.connect(self.toggle_calibration_fields)
        hydro_group_layout.addWidget(self.calibration_checkbox, 0, 1)

        # Auto-Calculate Vadose Zone Checkbox
        self.vadose_zone_checkbox = QCheckBox("Auto-Calculate Vadose Zone")
        self.vadose_zone_checkbox.setToolTip("Automatically calculate the vadose zone based on the mean vadose depth over an initial simulation")
        self.vadose_zone_checkbox.setChecked(True)
        self.vadose_zone_checkbox.setEnabled(False)
        self.vadose_zone_checkbox.toggled.connect(self.toggle_vadose_zone_fields)
        hydro_group_layout.addWidget(self.vadose_zone_checkbox, 0, 2, 1, 2)

        # Start and End Vadose Zone Fields (initially disabled)
        self.start_vadose_zone_label = QLabel("Start Vadose Zone (mm)")
        self.start_vadose_zone_input = QLineEdit()
        self.end_vadose_zone_label = QLabel("End Vadose Zone (mm)")
        self.end_vadose_zone_input = QLineEdit()
        self.start_vadose_zone_label.setEnabled(False)
        self.start_vadose_zone_input.setEnabled(False)
        self.end_vadose_zone_label.setEnabled(False)
        self.end_vadose_zone_input.setEnabled(False)

        # Layout for Vadose Zone fields
        hydro_group_layout.addWidget(self.start_vadose_zone_label, 1, 0)
        hydro_group_layout.addWidget(self.start_vadose_zone_input, 1, 1)
        hydro_group_layout.addWidget(self.end_vadose_zone_label, 1, 2)
        hydro_group_layout.addWidget(self.end_vadose_zone_input, 1, 3)

        # Calibration start and end fields
        self.calibration_start_label = QLabel("Calibration Start")
        self.calibration_start_label.setToolTip("Start date for the calibration period. Recommend YYYY1001 for hydrological year start.")
        self.calibration_start_input = QLineEdit()
        self.calibration_end_label = QLabel("Calibration End")
        self.calibration_end_label.setToolTip("End date for the calibration period. Recommend YYYY0930 for hydrological year end.")
        self.calibration_end_input = QLineEdit()
        self.calibration_start_label.setEnabled(False)
        self.calibration_start_input.setEnabled(False)
        self.calibration_end_label.setEnabled(False)
        self.calibration_end_input.setEnabled(False)
        hydro_group_layout.addWidget(self.calibration_start_label, 2, 0)
        hydro_group_layout.addWidget(self.calibration_start_input, 2, 1)
        hydro_group_layout.addWidget(self.calibration_end_label, 2, 2)
        hydro_group_layout.addWidget(self.calibration_end_input, 2, 3)

        #~~~~~~~~~~~~
        # XML Editor
        #~~~~~~~~~~~~
        # Load and parse the XML file
        parser = etree.XMLParser(remove_blank_text=True)
        tree = etree.parse(current_dir / 'resources' / 'HydroPlusConfig_CoolAir_Template.xml', parser)
        root = tree.getroot()

        #Title for XML Editor section
        xml_title_label = QLabel("XML Editor")
        xml_title_label.setStyleSheet("""
            QLabel {
                font-size: 12pt;
                font-weight: bold;
            }   
        """)
        xml_title_layout = QGridLayout()
        xml_title_layout.addWidget(xml_title_label, 0, 0, 1, 3)
        #Import XML Button
        self.xml_import_checkbox = QCheckBox("Import XML")
        self.xml_import_checkbox.setToolTip("Import an existing XML file from project set-up folder")
        xml_title_layout.addWidget(self.xml_import_checkbox, 0, 4)
        #XML checkbox aligned right

        # Create sections with recommended subsections wrapped in scrollable areas
        data_folder_group = self.create_data_folder_with_recommendations(root, 'DataOrganizer/DataDrawer/DataFolder',
                                                                     "DataFolder")
        temp_calc_params_group = self.create_temp_calc_with_recommendations(root, "TemperatureCalculationParams",
                                                                          "TemperatureCalculationParams")
        temp_execution_group = self.create_temp_execution_with_recommendations(root, "TemperatureExecutionParams", "TemperatureExecutionParams")
        #Textbox that explains stars on XML variables
        star_text = QLabel("*Calculated by the model\n** Calculated during Hydro calibration")
        # Wrap each group box in a scroll area and add to the tab layout
        scroll_area_data_folder = self.create_scrollable_group(data_folder_group)
        scroll_area_temp_calc_params = self.create_scrollable_group(temp_calc_params_group)
        scroll_area_temp_execution = self.create_scrollable_group(temp_execution_group)
        #Process Button
        process_box = QGroupBox("")
        process_layout = QGridLayout()
        process_box.setLayout(process_layout)
        self.process_button = QPushButton("Process")
        self.process_button.clicked.connect(self.process_data)
        process_layout.addWidget(self.process_button,0, 1, 1, 2)
        #Add widgets to tab2 layout
        tab2_layout.addLayout(adv_title_group_layout)
        tab2_layout.addWidget(advanced_group_box)
        tab2_layout.addWidget(hydro_group_box)
        tab2_layout.addLayout(xml_title_layout)
        tab2_layout.addWidget(scroll_area_data_folder)
        tab2_layout.addWidget(scroll_area_temp_calc_params)
        tab2_layout.addWidget(scroll_area_temp_execution)
        tab2_layout.addWidget(star_text)
        tab2_layout.addWidget(process_box)
        self.tab_widget.addTab(self.tab2, "Advanced Options")

        #===============================
        # TAB 3: Alternate Scenarios
        #===============================
        #Create scenario boxes to track number of scenarios
        self.scenario_boxes = []
        self.tab3 = QWidget()
        scenario_layout = QVBoxLayout(self.tab3)
        #Title
        scenario_label = QLabel("Alternate Scenarios")
        scenario_label.setStyleSheet("""
            QLabel {
                font-size: 16pt; 
            }
        """)
        scenario_label_layout = QGridLayout()
        scenario_label_layout.addWidget(scenario_label, 0, 0)

        #Help Button
        help_button = QPushButton("Help")
        help_button.clicked.connect(self.toggle_help_window_scenario)
        scenario_label_layout.addWidget(help_button, 0, 3)
        help_button.setMaximumWidth(50)

        scenario_layout.addLayout(scenario_label_layout)
        #Scenario generation
        for i in range(1, 4):
            #Layout
            temp_group_box = QGroupBox(f"Scenario {i}")
            temp_group_box.setStyleSheet("""
                QGroupBox {
                    font-size: 12pt;
                    font-weight: bold;
                }
            """)
            main_layout = QVBoxLayout(temp_group_box)
            #Scenario Name
            alias_box = QGroupBox("")
            alias_layout = QGridLayout()
            alias_box.setLayout(alias_layout)
            alias_label = QLabel("Scenario Name:")
            self.alias_input = QLineEdit(f"Alt_Case_{i}")
            alias_layout.addWidget(alias_label, 0, 0)
            alias_layout.addWidget(self.alias_input, 0, 1)
            main_layout.addWidget(alias_box)
            #'Where' Box
            where_group_box = QGroupBox("WHERE:")
            where_group_box.setStyleSheet("""
                QGroupBox {
                    font-size: 10pt;
                    font-weight: bold;
                }
            """)
            where_layout = QGridLayout()
            where_group_box.setLayout(where_layout)
            self.add_where_fields(where_layout)
            #'Action' Box
            action_group_box = QGroupBox("ACTION:")
            action_group_box.setStyleSheet("""
                QGroupBox {
                    font-size: 10pt;
                    font-weight: bold;
                }
            """)
            action_layout = QGridLayout()
            action_group_box.setLayout(action_layout)
            self.add_action_fields(action_layout)
            #Add where/action to main layout for each scenario
            main_layout.addWidget(where_group_box)
            main_layout.addWidget(action_group_box)
            #Add to scenario tab layout
            scenario_layout.addWidget(temp_group_box)
            #Add to list of scenario boxes
            self.scenario_boxes.append(temp_group_box)
        #Process Button
        process_box = QGroupBox("")
        process_layout = QGridLayout()
        process_box.setLayout(process_layout)
        self.process_button = QPushButton("Process")
        self.process_button.clicked.connect(self.process_data)
        process_layout.addWidget(self.process_button, 0, 1, 1, 2)
        scenario_layout.addWidget(process_box)
        self.tab_widget.addTab(self.tab3, "Scenarios")

        #===============================
        # TAB 4: Console P_ITCA_01
        #===============================
        self.tab4 = QWidget()
        tab4_layout = QVBoxLayout(self.tab4)

        # Console Window
        self.text_edit = QTextEdit()
        self.text_edit.setReadOnly(True)

        # Run Cool Air Button
        self.run_button = QPushButton("Run Cool Air")
        self.run_button.clicked.connect(self.run_batches)
        self.run_button.setEnabled(False)

        # Terminate Button
        self.terminate_button = QPushButton("Terminate")
        self.terminate_button.clicked.connect(self.terminate_script)
        self.terminate_button.setEnabled(False)  # Initially disabled

        # Layout for buttons
        button_layout = QGridLayout()
        button_layout.addWidget(self.terminate_button, 0, 0, 1, 1)

        button_layout.addWidget(self.run_button, 0, 1, 1, 3)

        tab4_layout.addWidget(self.text_edit)
        tab4_layout.addLayout(button_layout)

        self.tab_widget.addTab(self.tab4, "Console")
        self.tab_widget.setTabEnabled(3, False)

        #===============================
        # TAB 5: Mosaic Outputs P_ITCA_03
        #===============================
        self.tab5 = QWidget()
        tab5_layout = QVBoxLayout(self.tab5)

        #Units Outputs
        self.units_output = QGroupBox("Temperature Units")
        units_layout = QGridLayout()
        self.units_output.setLayout(units_layout)
        self.units_output.setStyleSheet("""
            QGroupBox {
                font-size: 10pt;
                font-weight: bold;
            }
        """)
        self.kelvin_checkbox = QCheckBox("Kelvin")
        self.kelvin_checkbox.setChecked(True)
        self.fahrenheit_checkbox = QCheckBox("Fahrenheit")
        self.fahrenheit_checkbox.setChecked(True)
        self.celsius_checkbox = QCheckBox("Celsius")

        units_layout.addWidget(self.kelvin_checkbox, 0, 0)
        units_layout.addWidget(self.fahrenheit_checkbox, 0, 1)
        units_layout.addWidget(self.celsius_checkbox, 0, 2)


        #Auto-run button
        self.auto_run_mosaic_button = QPushButton("Auto-Run Output Maps")
        self.auto_run_mosaic_button.clicked.connect(self.auto_run_mosaic)
        self.auto_run_mosaic_button.setEnabled(False)

        #Or Manual Input
        mosaic_start_label = QLabel("Mosaic Start Date:")
        self.mosaic_start_input = QLineEdit("")
        mosaic_end_label = QLabel("Mosaic End Date:")
        self.mosaic_end_input = QLineEdit("")
        mosaic_epsg_label = QLabel("Output EPSG:")
        self.mosaic_epsg_input = QLineEdit("5070")
        mosaic_folder_label = QLabel("Mosaic Folder:")
        self.mosaic_folder_output = QLineEdit("")
        mosaic_button = QPushButton("Browse...")
        mosaic_button.clicked.connect(lambda: self.open_folder_dialog(self.mosaic_folder_output))
        self.process_mosaics_button = QPushButton("Process Output Maps")
        self.process_mosaics_button.clicked.connect(self.manual_run_mosaic)

        auto_run_group = QGroupBox("Auto-Run Mosaic Output")
        auto_run_layout = QGridLayout()
        auto_run_group.setLayout(auto_run_layout)
        auto_run_group.setStyleSheet("""
            QGroupBox {
                font-size: 10pt;
                font-weight: bold;
                }
        """)

        #Add widgets to layout
        auto_run_layout.addWidget(self.auto_run_mosaic_button, 0, 0, 1, 2)

        manual_group = QGroupBox("Manual Mosaic Output")
        manual_layout = QGridLayout()
        manual_group.setLayout(manual_layout)
        manual_group.setStyleSheet("""
            QGroupBox {
                font-size: 10pt;
                font-weight: bold;
                }
            """)
        manual_layout.addWidget(mosaic_start_label, 0, 0)
        manual_layout.addWidget(self.mosaic_start_input, 0, 1)
        manual_layout.addWidget(mosaic_end_label, 1, 0)
        manual_layout.addWidget(self.mosaic_end_input, 1, 1)
        manual_layout.addWidget(mosaic_epsg_label, 0, 2)
        manual_layout.addWidget(self.mosaic_epsg_input, 0, 3)
        manual_layout.addWidget(mosaic_folder_label, 2, 0)
        manual_layout.addWidget(self.mosaic_folder_output, 2, 1, 1, 2)
        manual_layout.addWidget(mosaic_button, 2, 3)
        manual_layout.addWidget(self.process_mosaics_button, 3, 0, 1, 4)

        #Console window for mosaic
        self.text_edit_mosaic = QTextEdit()
        self.text_edit_mosaic.setReadOnly(True)
        self.text_edit_mosaic.setText("================================\n"
                                      "Mosaic Processor Console\n"
                                      "================================\n")
        tab5_layout.addWidget(self.units_output)
        tab5_layout.addWidget(auto_run_group)
        tab5_layout.addWidget(manual_group)
        tab5_layout.addWidget(self.text_edit_mosaic)
        self.tab_widget.addTab(self.tab5, "Mosaic Outputs")


if __name__ == "__main__":
    import sys

    app = QApplication(sys.argv)
    splash = SplashScreen()
    splash.show()

    window = MainWindow()
    splash.finish(window)
    window.show()  # Show the main window
    sys.exit(app.exec())  # Execute the application's main loop
