import os
import geopandas as gpd
import rasterio
from osgeo import gdal
from netCDF4 import Dataset
from rasterio.mask import mask
from rasterio.warp import calculate_default_transform, reproject, Resampling
from rasterio.features import rasterize
from rasterio.vrt import WarpedVRT
from shapely.geometry import mapping, shape, box
from shapely.ops import transform
from pyproj import CRS, Transformer
from rasterio.io import MemoryFile
from rasterio.transform import from_origin, from_bounds, Affine, array_bounds
from pathlib import Path
import fiona
import pyproj
import xml.etree.ElementTree as ET
import numpy as np
import shutil
import argparse
import json
import csv
import subprocess
import sys
gdal.UseExceptions()
#TO-DO
#Add in AH4GUC processing
#Add in field for DEM reference and remove process weather option
#

#Folder Creation
def create_project_folders(folder):
    if not folder.exists():
        folder.mkdir(parents=True, exist_ok=True)
#Script 1: Clip raster to buffered AOI

def calculate_bounding_box(input_raster, input_aoi_shapefile):
    # Read the input raster
    with rasterio.open(input_raster) as src:
        raster_crs = src.crs

    # Read the input AOI shapefile
    aoi_gdf = gpd.read_file(input_aoi_shapefile)
    aoi_crs = aoi_gdf.crs

    # Reproject the AOI shapefile to match the raster's CRS
    aoi_gdf = aoi_gdf.to_crs(raster_crs)

    # Get the geometry of the AOI
    aoi_geometry = aoi_gdf.geometry[0]

    # Calculate the bounding box of the AOI with a 10% buffer
    minx, miny, maxx, maxy = aoi_geometry.bounds
    buffer_size = max(maxx - minx, maxy - miny) * 0.1
    buffered_bbox = box(minx - buffer_size, miny - buffer_size,
                        maxx + buffer_size, maxy + buffer_size)

    return buffered_bbox, raster_crs

def rasterize_polygons(filepath_shape, bounding_box, bounding_box_crs, output_tif, out_res):
    # Read the shapefile
    shape_gdf = gpd.read_file(filepath_shape)

    # Create a new 'FID' column corresponding to the index
    shape_gdf['FID'] = shape_gdf.index


    # Reproject the shapefile to match the bounding box's CRS if necessary
    if shape_gdf.crs != bounding_box_crs:
        shape_gdf = shape_gdf.to_crs(bounding_box_crs)

    # Create a GeoDataFrame for the bounding box with its CRS
    bounding_box_gdf = gpd.GeoDataFrame({'geometry': [bounding_box]}, crs=bounding_box_crs)
    # Clip the shapefile data to the bounding box
    shape_clipped = gpd.clip(shape_gdf, bounding_box_gdf)
    # Check if shape_clipped is empty
    if shape_clipped.empty:
        print("Warning: No features found within the bounding box after clipping.")
        print("Please check if the bounding box and shapefile overlap.")
        return

    # Get the bounds of the bounding box in the bounding box's CRS
    minx, miny, maxx, maxy = bounding_box_gdf.total_bounds

    # Convert to floats to ensure proper division
    minx = float(minx)
    maxx = float(maxx)
    miny = float(miny)
    maxy = float(maxy)
    pixel_size = float(out_res)

    # Calculate the number of pixels in each dimension
    width = int(np.ceil((maxx - minx) / pixel_size))
    height = int(np.ceil((maxy - miny) / pixel_size))

    # Define the transform for the rasterization
    transform = from_origin(minx, maxy, pixel_size, pixel_size)

    # Prepare the shapes and their corresponding values using the 'FID' field
    shapes = ((geom, fid) for geom, fid in zip(shape_clipped.geometry, shape_clipped['FID']))

    # Rasterize the shapes
    try:
        raster_array = rasterize(
            shapes=shapes,
            out_shape=(height, width),
            transform=transform,
            fill=0,
            all_touched=False,
            dtype='int32'
        )
    except Exception as e:
        return

    unique_values = np.unique(raster_array)

    # Define the metadata for the output raster
    out_meta = {
        'driver': 'GTiff',
        'height': height,
        'width': width,
        'count': 1,
        'dtype': 'int32',
        'crs': bounding_box_crs,
        'transform': transform
    }

    # Write the raster to the output TIF file
    with rasterio.open(output_tif, 'w', **out_meta) as dest:
        dest.write(raster_array, 1)

def netcdf_to_tif(filepath_ah4guc, month, hour, out_res, output_tif, bounding_box, bounding_box_crs):
    # Open the NetCDF file in read-only mode
    nc = Dataset(filepath_ah4guc, 'r')

    # Get the data variable 'ahe'
    ahe = nc.variables['ahe']

    # Get the indices for the given month and hour
    month_idx = month
    hour_idx = hour

    # Extract the data for the given month and hour
    data = ahe[month_idx, hour_idx, :, :]

    # Convert data to float32
    data = data.astype(np.float32)

    # Handle masked arrays
    if isinstance(data, np.ma.MaskedArray):
        data = data.filled(np.nan)

    # Flip the data array vertically to match the coordinate system
    data = np.flipud(data)

    # Get the latitudes and longitudes
    lats = nc.variables['lat'][:]
    lons = nc.variables['lon'][:]

    # Close the NetCDF file
    nc.close()

    # Ensure latitudes are increasing from south to north
    if lats[0] > lats[-1]:
        lats = lats[::-1]
        # No need to flip data again since we've already done it


    # Create the source transform using from_bounds
    min_lon = lons.min()
    max_lon = lons.max()
    min_lat = lats.min()
    max_lat = lats.max()
    src_transform = from_bounds(min_lon, min_lat, max_lon, max_lat, data.shape[1], data.shape[0])

    # Define the source CRS (assuming WGS84 Latitude/Longitude)
    src_crs = CRS.from_epsg(4326)

    # Define the destination CRS (bounding_box_crs)
    dst_crs = bounding_box_crs

    # Calculate the transform and dimensions for the reprojected data
    dst_transform, dst_width, dst_height = calculate_default_transform(
        src_crs, dst_crs, data.shape[1], data.shape[0],
        left=min_lon, bottom=min_lat, right=max_lon, top=max_lat
    )

    reprojected_data = np.empty((dst_height, dst_width), dtype=np.float32)
    reproject(
        source=data,
        destination=reprojected_data,
        src_transform=src_transform,
        src_crs=src_crs,
        dst_transform=dst_transform,
        dst_crs=dst_crs,
        resampling=Resampling.nearest,
        src_nodata=np.nan,
        dst_nodata=np.nan
    )

    # Clip the data to the bounding box
    with MemoryFile() as memfile:
        with memfile.open(
            driver='GTiff',
            height=reprojected_data.shape[0],
            width=reprojected_data.shape[1],
            count=1,
            dtype=reprojected_data.dtype,
            crs=dst_crs,
            transform=dst_transform,
            nodata=np.nan
        ) as dataset:
            dataset.write(reprojected_data, 1)
            # Create a geometry mapping for the bounding box
            bbox_geom = [mapping(bounding_box)]
            # Mask the data with the bounding box and crop it
            out_image, out_transform = mask(
                dataset,
                bbox_geom,
                crop=True,
                nodata=np.nan
            )

    # Handle masked arrays
    if isinstance(out_image, np.ma.MaskedArray):
        out_image = out_image.filled(np.nan)

    # Calculate the bounds of the clipped data
    bounds = array_bounds(out_image.shape[1], out_image.shape[2], out_transform)

    # Calculate new transform and dimensions based on out_res
    new_transform, new_width, new_height = calculate_default_transform(
        dst_crs, dst_crs, out_image.shape[2], out_image.shape[1],
        left=bounds[0], bottom=bounds[1], right=bounds[2], top=bounds[3],
        resolution=out_res
    )

    resampled_data = np.empty((new_height, new_width), dtype=np.float32)
    reproject(
        source=out_image,
        destination=resampled_data,
        src_transform=out_transform,
        src_crs=dst_crs,
        dst_transform=new_transform,
        dst_crs=dst_crs,
        resampling=Resampling.bilinear,
        src_nodata=np.nan,
        dst_nodata=np.nan
    )

    # Define the metadata for the output raster
    out_meta = {
        "driver": "GTiff",
        "height": resampled_data.shape[0],
        "width": resampled_data.shape[1],
        "count": 1,
        "dtype": resampled_data.dtype,
        "crs": dst_crs,
        "transform": new_transform,
        "nodata": np.nan
    }

    # Write the resampled data to the output GeoTIFF file
    with rasterio.open(output_tif, "w", **out_meta) as dest:
        dest.write(resampled_data, 1)

def read_raster(raster_path):
    with rasterio.open(raster_path) as src:
        return src.read(1), src.nodatavals[0]

def calculate_cool_air_frac(TreeCover_array_percent,ImperviousCover_array_percent,LandCover_array_class):
    def calculate_statistics(ic_data, tc_data):
        """
        Calculates average IC under TC for i-Tree Cool Air simulations
        """
        # Apply mask to filter valid data
        mask = (tc_data >= 1) & (tc_data <= 100)
        ic_clipped = ic_data[mask]
        tc_clipped = tc_data[mask]

        # Get unique tc_values and their counts
        unique_tc_values, inverse_indices = np.unique(tc_clipped, return_inverse=True)
        tc_counts = np.bincount(inverse_indices)

        # Calculate mean ic_values for each unique tc_value
        ic_sums = np.bincount(inverse_indices, weights=ic_clipped)
        ic_means = ic_sums / tc_counts

        # Prepare the stats dictionary
        stats = {
            'tc_value': unique_tc_values,
            'ic_mean': ic_means,
            'tc_count': tc_counts
        }

        return stats
    # Create a mask for developed areas (classes 21, 22, 23, 24)
    mask_developed = np.isin(LandCover_array_class, [21, 22, 23, 24])

    # Clip TreeCover and ImperviousCover arrays to developed areas
    tc_clipped = TreeCover_array_percent[mask_developed]
    ic_clipped = ImperviousCover_array_percent[mask_developed]

    # Calculate statistics on the masked IC and TC data
    stats_df = calculate_statistics(ic_clipped, tc_clipped)
    tc_count_redc = np.nansum(stats_df['tc_count'])

    # Normalize and weight the IC values based on the TC counts
    stats_df['normalized_tc_count'] = stats_df['tc_count'] / tc_count_redc
    stats_df['weighted_ic'] = stats_df['ic_mean'] * stats_df['normalized_tc_count']

    # Calculate the total weighted impervious cover
    total_weighted_ic = stats_df['weighted_ic'].sum()

    # Calculate TreeCover_overImperviousCover_CoolAir
    TreeCover_overImperviousCover_CoolAir = total_weighted_ic

    return TreeCover_overImperviousCover_CoolAir

def write_ascii_header(ascii_file, header_info):
    """
    Write ASCII header information to the specified file.
    """
    for key, value in header_info.items():
        ascii_file.write(f'{key.ljust(13)}{value}\n')

def update_xml(parent, child_tag, new_text):
    """
    Update the text of a child element in an XML tree.

    Parameters:
    parent (Element): The parent element.
    --Singular:root.find('SimulationStringParams')
    --Multiple: root.find('DataOrganizer/DataDrawer/DataFolder')
    child_tag (str): The tag of the child element to update.
    new_text (str): The new text to set for the child element.

    Dependencies:
    import xml.etree.ElementTree as ET

    Example Code:
    # Load and parse the XML file
    tree = ET.parse(destination_xml)
    root = tree.getroot()

    data_folder = root.find('DataOrganizer/DataDrawer/DataFolder')
        if data_folder is not None:
            update_xml(data_folder, 'Area_m2', total_area)

    tree.write(destination_xml)
    """
    child = parent.find(child_tag)
    if child is not None:
        child.text = str(new_text)

def delete_xml(parent, child_tag):
    """
    Remove a child element from a parent element in an XML tree.

    Parameters:
    parent (Element): The parent element.
    child_tag (str): The tag of the child element to remove.

    Example:
    remove_xml_element(temperature_execution_params, 'PrintBlockGroupRange0')
    """
    child = parent.find(child_tag)
    if child is not None:
        parent.remove(child)

def create_batch_file(folder, batch_num, quads_list, file_prefix):
    batch_file_path = folder / f'{file_prefix}-{batch_num}.bat'
    with batch_file_path.open("w") as batch_file:
        for quad_folder in quads_list:
            batch_file.write(f"HydroPlus.exe {quad_folder}\n")
        print(f"Batch file {file_prefix}-{batch_num}.bat created.")

def get_hydroplus_config_path():
    # Get the directory of the current script
    script_dir = Path(__file__).resolve().parent.parent

    # Construct the path to the HydroPlus_Config_Template.xml file
    hydroplus_config_template = script_dir / 'resources' / 'HydroPlusConfig_CoolAir_Template.xml'

    return hydroplus_config_template
#=========================
#   USER INPUTS
#=========================

def main(args):
    from pathlib import Path
    #Load in JSON Configuration file
    print("Loading config file...")
    with open(args.config_filename, 'r') as json_file:
        content = json_file.read()  # Read content to check if it's correctly accessible
        config_data = json.loads(content)
    print(f"Configuration for project {config_data['proj_name']} loaded successfully.")

    #Project name
    proj_name = config_data['proj_name']

    #Raster resolution for input and outputs
    input_resolution = 30
    output_resolution_list = config_data['output_resolution_list']
    #flag_NumberofBatches: Number of batches to be run simultaneously. Default is 1. Use caution over 8.
    flag_NumberofBatches = config_data['flag_NumberofBatches']
    #flag_OutputEPSG: EPSG code for output rasters
    #5070 = NAD_1983_Contiguous_USA_Albers
    #326[##] = WGS 84 / UTM zone ##N
    #327[##] = WGS 84 / UTM zone ##S
    flag_OutputEPSG = config_data['flag_OutputEPSG']
    #Path to Weather and Radiation.csv
    path_weather = config_data['path_weather']
    #Pathlib notation

    #flag_MultipleStation: Use multiple weather stations
    flag_MultipleStation = config_data['flag_MultipleStation']
    if flag_MultipleStation:
        #Path to WeatherStationAttributeTable.csv
        filepath_multistation = config_data['filepath_multistation']
        #flag_HottestDaySimulated: Choose hottest day to simulate regardless of SimulationStart/End. Only for use with multiple stations.
        flag_HottestDaySimulated = config_data['flag_HottestDaySimulated']

    #flag_NLCDAutoSearch: Automatically search for NLCD files based on prefix nlcd_yyyy_{file type}_l48
    #0=Disabled; manually define file names below (file_tc, file_ic, file_lc), #1 = Enabled
    flag_NLCDAutoSearch = config_data['flag_NLCDAutoSearch']
    if flag_NLCDAutoSearch:
        #flag_NLCD_LC_year: Year of NLCD data
        #2016, 2019, or 2021. Used for auto-search only.
        flag_NLCD_LC_year = config_data['flag_NLCD_LC_year']
    else:
        file_tc = config_data['file_tc']
        file_ic = config_data['file_ic']
        file_lc = config_data['file_lc']


    #Change: All paths
    #path_trunk: location of z_places folder created to hold intermediate inputs
    path_trunk = config_data['path_trunk']
    #flag_XML_Import: Enable XML import for HydroPlusConfig.xml, will search in path_trunk + proj_name_name
    flag_XML_Import = config_data['flag_XML_Import']

    #path_project: location of project folder where Cool Air runs will be stored
    path_project = config_data['path_project']

    #path_aoi: path for AOI polygon
    filepath_aoi = config_data['filepath_aoi']

    #path_dem and file_dem for path and filename for DEM
    filepath_dem = config_data['filepath_dem']
    #flag_DEM_ratio: Conversion ratio for DEM.
    #Input DEM in meters = 1, Input DEM in centimeters = 100
    #100 for raw NHD Data
    flag_DEM_ratio = config_data['flag_DEM_ratio']

    #path_nlcd and file_nlcd
    path_nlcd = config_data['path_nlcd']

    #Anthropogenic heat flag and filepath
    flag_AnthropogenicHeat = config_data['flag_AnthropogenicHeat']
    if flag_AnthropogenicHeat:
        filepath_ah4guc = config_data['filepath_ah4guc']
    else:
        filepath_ah4guc = ""

    #Census / Block Group optional input
    try:
        flag_BlockGroup = config_data['flag_Census']
    except KeyError:
        #Handle old versions of JSON that did not include this flag yet
        flag_BlockGroup = False
    if flag_BlockGroup:
        filepath_blockgroup = config_data['filepath_census']
        census_fids = config_data['census_fids']
    else:
        filepath_blockgroup = ""

    #Script 2
    #Maximum number of rows/cols for each i-Tree Cool Air Simulation; RAM dependent. Upper limit of 300x300.
    col_max = int(config_data['col_max'])
    row_max = int(config_data['row_max'])
    simulation_start = config_data['simulation_start']
    simulation_end = config_data['simulation_end']

    #Retrieve parent folder of script
    hydroplus_config_template = get_hydroplus_config_path()
    #flag_AdvancedConfiguration: Enable custom variable inputs into HydroPlusConfig.xml
    xml_import = config_data['xml_import']
    flag_DEM_average = config_data['flag_DEM_average']
    #These values will always be changed/updated
    temperature_execution_config = xml_import['TemperatureExecutionParams']
    data_folder_config = xml_import['DataFolder']
    temperature_calculation_config = xml_import['TemperatureCalculationParams']

    #flag_Hydro: Enable Hydro processing
    flag_Hydro = config_data['flag_Hydro']
    if flag_Hydro:
        hydro_project_path = config_data['hydro_path']
        flag_Calibration = config_data['flag_Calibration']
        if flag_Calibration:
            calibration_start = config_data['calibration_start']
            calibration_end = config_data['calibration_end']
            flag_Vadose_auto = config_data['flag_Vadose_auto']
            if not flag_Vadose_auto:
                first_vadose = config_data['first_vadose']
                last_vadose = config_data['last_vadose']
            else:
                first_vadose = ''
                last_vadose = ''
    else:
        flag_Calibration = False


    #Scenarios
    if config_data['all_scenarios_data'] is not None:
        scenarios = config_data['all_scenarios_data']

    # Create batch list to send back to CoolAir.py
    batch_files = []
    #Create output_folder list to send back to CoolAir.py
    output_folders_list = []

    #Print metadata
    print("=====================\n"
          "PROJECT SET-UP\n"
          "=====================\n"
          f"proj_name = {proj_name}\n"
          f"input_resolution = {input_resolution}\n"
          f"output_resolution = {output_resolution_list}\n", flush=True)
    #Weather Info
    #Print USAF-WBAN if using WeatherPrep
    #Land Cover Info
    print("==========\n"
          "Land Cover\n"
          "==========")
    print(f"flag_AnthropogenicHeat = {flag_AnthropogenicHeat}")
    if flag_AnthropogenicHeat:
        print(f"filepath_ah4guc: {filepath_ah4guc}")
    print(f"flag_Census = {flag_BlockGroup}")
    if flag_BlockGroup:
        print(f"filepath_census: {filepath_blockgroup}\n"
              f"census_fids: {census_fids}")
    print(f"flag_NLCDAutoSearch = {flag_NLCDAutoSearch}")
    if flag_NLCDAutoSearch == 0:
        print(f"file_lc = {path_nlcd}{file_lc}\n"
              f"file_tc = {path_nlcd}{file_tc}\n"
              f"file_ic = {path_nlcd}{file_ic}")
    else:
        print(f"flag_NLCD_LC_year = {flag_NLCD_LC_year}")
    #Folder paths
    print(f"trunk folder = {path_trunk}\n"
          f"shapefile location = {filepath_aoi}\n"
          f"nlcd location = {path_nlcd}\n"
          f"dem location = {filepath_dem}")
    #DEM Info
    if flag_DEM_ratio == 100:
        print("input dem units: centimeters")
    elif flag_DEM_ratio == 1:
        print("input dem units: meters")
    else:
        print(f"custom DEM ratio: {flag_DEM_ratio}")
    #Script 2 Config
    print("=====================\n"
          "Cool Air Configuration\n"
          "=====================")
    print(f"col_max = {col_max}\n"
          f"row_max = {row_max}\n"
          f"simulation_start = {simulation_start}\n"
          f"simulation_end = {simulation_end}\n"
          f"temperature_execution_config = {temperature_execution_config}\n"
          f"data_folder_config = {data_folder_config}\n"
          f"temperature_calculation_config = {temperature_calculation_config}\n"
          f"flag_DEM_average = {flag_DEM_average}\n"
          f"flag_NumberofBatches = {flag_NumberofBatches}\n")
    print("=====================\n"
          "Hydro Configuration\n"
          "=====================")
    print(f"flag_Hydro = {flag_Hydro}\n"
          f"flag_Calibration = {flag_Calibration}")
    if flag_Calibration:
        print(f"calibration_start = {calibration_start}\n"
              f"calibration_end = {calibration_end}\n")
        if not flag_Vadose_auto:
            print(f"first_vadose = {first_vadose}\n"
                  f"last_vadose = {last_vadose}")
        else:
            print("flag_Vadose_auto = True")


    for out_res in output_resolution_list:
        print("1: Creating project folders...")
        # Create project folders
        create_project_folders(Path(path_project) / proj_name)
        create_project_folders(Path(path_trunk) / proj_name / f'res_{str(out_res)}')
        create_project_folders(Path(path_project) / proj_name / f'Inputs{str(out_res)}')
        create_project_folders(Path(path_project) / proj_name / f'Outputs{str(out_res)}')
        output_folders_list.append(Path(path_project) / proj_name / f'Outputs{str(out_res)}')
        script1_outfolder = Path(path_trunk) / proj_name / f'res_{str(out_res)}'

        #Initialize filepaths for rasters
        filepath_tc = ""
        filepath_ic = ""
        filepath_lc = ""
        if flag_NLCDAutoSearch:
            nlcd_path = Path(path_nlcd)
            for file in nlcd_path.iterdir():
                if file.stem.startswith(f"NLCD_{flag_NLCD_LC_year}_Tree_Canopy_L48") and file.suffix in {".tiff", ".tif"}:
                    filepath_tc = nlcd_path / file
                # New tree canopy file naming convention
                elif file.stem.startswith(f"nlcd_tcc_conus_{flag_NLCD_LC_year}") and file.suffix in {".tiff", ".tif"}:
                    filepath_tc = nlcd_path / file
                elif file.stem.startswith(f"NLCD_{flag_NLCD_LC_year}_Impervious_L48") and file.suffix in {".tiff", ".tif"}:
                    filepath_ic = nlcd_path / file
                elif file.stem.startswith(f"NLCD_{flag_NLCD_LC_year}_Land_Cover_L48") and file.suffix in {".tiff", ".tif"}:
                    filepath_lc = nlcd_path / file
            if flag_NLCD_LC_year == 2019 and filepath_tc == "":
                nlcd_path = Path(path_nlcd)
                for file in nlcd_path.iterdir():
                    if file.stem.startswith(f"NLCD_2016_Tree_Canopy_L48") and file.suffix in {".tiff", ".tif"}:
                        filepath_tc = nlcd_path / file
        else:
            nlcd_path = Path(path_nlcd)
            filepath_tc = f"{nlcd_path}/{file_tc}"
            filepath_ic = f"{nlcd_path}/{file_ic}"
            filepath_lc = f"{nlcd_path}/{file_lc}"

        #Project filepath_aoi to filepath_lc & calculate bounding box with 10% buffer for clipping before further processing
        print(f"2: Calculating project bounds from {filepath_aoi}")
        bounding_box, bounding_box_crs = calculate_bounding_box(filepath_lc, filepath_aoi)
        #Open AOI shapefile and project to flag_OutputEPSG
        aoi_gdf = gpd.read_file(filepath_aoi)
        aoi_gdf = aoi_gdf.to_crs(flag_OutputEPSG)
        output_aoi = Path(path_trunk) / proj_name / f'boundary-epsg{flag_OutputEPSG}.shp'
        #Save AOI shapefile to path_trunk
        print(f"3: Saving AOI shapefile to {output_aoi}")
        aoi_gdf.to_file(output_aoi)
        #Make list of rasters to process, changed depending on AH / Census flags
        raster_list = [filepath_lc, filepath_tc, filepath_ic, filepath_dem]
        #List to intermediate rasters created from optional inputs - deleted after processing
        optional_tmp = []
        if flag_AnthropogenicHeat:
            print("3b: Pre-processing Anthropogenic Heat...")
            #Find month of SimulationStart, extract MM from YYYYMMDDHH. Subtract 1 for 0-based index
            month = int(simulation_start[4:6]) - 1
            #Defualt hour to 20 for hottest hour
            hour=20
            filepath_ahe = str(script1_outfolder / f'ah4guc_{proj_name}_{out_res}_{month}{hour}.tif')
            #Check if filepath_ah4guc ends in .nc or .tif
            if filepath_ah4guc.endswith('.nc'):
                netcdf_to_tif(filepath_ah4guc, month, hour, out_res, filepath_ahe, bounding_box, bounding_box_crs)
                optional_tmp.append(filepath_ahe)
            else:
                #Overwrite filepath_ahe with filepath_ah4guc
                filepath_ahe = filepath_ah4guc
            raster_list.append(filepath_ahe)
        else:
            filepath_ahe = ""
        if flag_BlockGroup:
            print("3b: Pre-processing Census Data...")
            filepath_bg = str(script1_outfolder / 'blockgroup.tif')
            rasterize_polygons(filepath_blockgroup, bounding_box, bounding_box_crs, filepath_bg, out_res)
            raster_list.append(filepath_bg)
            optional_tmp.append(filepath_bg)
        else:
            filepath_bg = ""
        # Define filepaths and naming for each raster type
        for filepath_raster in raster_list:
            if filepath_raster == filepath_tc:
                script1_tif = script1_outfolder / f'tc_{proj_name}_{out_res}.tif'
                script1_tmp = script1_outfolder / f'tc_{proj_name}_{out_res}_tmp.tif'
                script2_asc = 'treecover.asc'
                file_descriptor = 'Tree Canopy'
                first_raster = False
            elif filepath_raster == filepath_ic:
                script1_tif = script1_outfolder / f'ic_{proj_name}_{out_res}.tif'
                script1_tmp = script1_outfolder / f'ic_{proj_name}_{out_res}_tmp.tif'
                script2_asc = 'imperviouscover.asc'
                file_descriptor = 'Impervious Cover'
                first_raster = False
            elif filepath_raster == filepath_lc:
                script1_tif = script1_outfolder / f'lc_{proj_name}_{out_res}.tif'
                script2_asc = 'landcover.asc'
                file_descriptor = 'Land Cover'
                first_raster = True
            elif filepath_raster == filepath_dem:
                script1_tif = script1_outfolder / f'dem_{proj_name}_{out_res}.tif'
                script1_tmp = script1_outfolder / f'dem_{proj_name}_{out_res}_tmp.tif'
                script2_asc = 'dem.asc'
                file_descriptor = 'DEM'
                first_raster = False
            if filepath_ahe != "":
                if filepath_raster == filepath_ahe:
                    script1_tif = script1_outfolder / f'ah4guc_{proj_name}_{out_res}.tif'
                    script1_tmp = script1_outfolder / f'ah4guc_{proj_name}_{out_res}_tmp.tif'
                    script2_asc = 'ah_max_w_p_m2.asc'
                    file_descriptor = 'Anthropogenic Heat'
                    first_raster = False
            if filepath_bg != "":
                if filepath_raster == filepath_bg:
                    script1_tif = script1_outfolder / f'blockgroup_{proj_name}_{out_res}.tif'
                    script1_tmp = script1_outfolder / f'blockgroup_{proj_name}_{out_res}_tmp.tif'
                    script2_asc = 'blockgroup.asc'
                    file_descriptor = 'Census'
                    first_raster = False

            output_crs = CRS.from_epsg(flag_OutputEPSG)
            #SCRIPT 1
            with rasterio.open(filepath_raster) as src:
                print(f"=====================================================\n"
                      f"4: Raster Processing: {file_descriptor}..."
                      f"\n=====================================================")
                # Mask with bounding box
                print(f"4b: Clipping raster to bounding box...")
                out_image, out_transform = mask(src, [bounding_box], crop=True)
                out_meta = src.meta.copy()

                if src.dtypes[0] == 'uint16':
                    out_meta.update(dtype='int16')
                    out_image = out_image.astype('int16')
                if src.dtypes[0] == 'uint8':
                    # Set any value greater than 100 to -9999
                    out_image = np.where(out_image > 100, np.array(-9999).astype(src.dtypes[0]), out_image)
                    out_meta.update(dtype='int16')
                    out_image = out_image.astype('int16')

                # Update the metadata with the new dimensions and transform
                out_meta.update({
                    "driver": "GTiff",
                    "height": out_image.shape[1],
                    "width": out_image.shape[2],
                    "transform": out_transform,
                    "nodata": -9999
                })
                # Create an in-memory file object
                with MemoryFile() as memfile:
                    # Create a new dataset in the memory file with the clipped data
                    with memfile.open(**out_meta) as dest:
                        dest.write(out_image)

                    # Rewind the memory file to the beginning
                    memfile.seek(0)
                    # Open the in-memory clipped raster
                    with memfile.open() as src:
                        src_array = src.read(1)
                        transform, width, height = calculate_default_transform(
                            src.crs, output_crs, src.width, src.height, *src.bounds, resolution=out_res
                        )

                        kwargs = src.meta.copy()

                        kwargs.update({
                            'transform': transform,
                            'width': width,
                            'height': height,
                            'nodata': -9999
                        })

                        # Create a new in-memory file for the projected raster
                        print("4c: Reprojecting raster to output EPSG...")
                        with MemoryFile() as projected_memfile:
                            # Create a new dataset in the projected memory file
                            with projected_memfile.open(**kwargs) as dest:
                                # Process the raster data in chunks
                                for i in range(1, src.count + 1):
                                    src_array = src.read(i)

                                    reproject(
                                        source=rasterio.band(src, i),
                                        destination=rasterio.band(dest, i),
                                        src_transform=src.transform,
                                        src_crs=src.crs,
                                        dst_transform=transform,
                                        dst_crs=output_crs,
                                        resampling=Resampling.nearest
                                    )

                                    # Write the reprojected chunk to the destination dataset
                                    dest.write(src_array, i)

                            # Rewind the projected memory file to the beginning
                            projected_memfile.seek(0)
                            # Perform further processing with the projected raster in memory...

                          #  print("Projected raster is now available in memory.")
                            with projected_memfile.open() as src:
                                # Get the CRS of the extent raster
                                extent_crs = src.crs

                                # Read the AOI shapefile
                                with fiona.open(output_aoi, 'r') as shapefile:
                                    shapes = [feature['geometry'] for feature in shapefile]
                                    aoi_crs = shapefile.crs
                                print(f"4d: Masking raster with AOI...")
                                # Mask the extent raster with the AOI shapefile1
                                out_image, out_transform = rasterio.mask.mask(src, shapes, crop=True)
                                out_meta = src.meta.copy()


                                # Update the metadata with the new dimensions and transform
                                out_meta.update({
                                    'driver': 'GTiff',
                                    'height': out_image.shape[1],
                                    'width': out_image.shape[2],
                                    'transform': out_transform,
                                    'crs': aoi_crs
                                })

                                # Check if current raster is the DEM and needs adjustment
                                if filepath_raster == filepath_dem and flag_DEM_ratio != 1:
                                    print("4e: Adjusting DEM values based on flag_DEM_ratio...")
                                    flag_DEM_ratio = float(flag_DEM_ratio)
                                    # Create a masked array where nodata values are ignored
                                    nodata = out_meta.get('nodata')
                                    if nodata is not None:
                                        mask_array = out_image == nodata
                                        safe_image = np.where(mask_array, out_image, out_image / flag_DEM_ratio)
                                    else:
                                        #Update nodata to -9999
                                        out_meta.update({
                                            'nodata': -9999
                                        })
                                        nodata = out_meta.get('nodata')
                                        mask_array = out_image == nodata
                                        safe_image = np.where(mask_array, out_image, out_image / flag_DEM_ratio)

                                    out_image = safe_image
                                    if filepath_raster != filepath_bg:
                                        out_meta.update({
                                            'dtype': 'float32',
                                        })  # Use the adjusted raster data
                                # Create a new raster dataset with the updated metadata and CRS
                                print("4f: Saving masked raster to file...")
                                with rasterio.open(script1_tif, 'w', **out_meta) as dest:
                                    dest.write(out_image)
                                    # Close the raster dataset
                                    dest.close()
                                # Reopen the masked raster to check the CRS
                                current_ds = gdal.Open(str(script1_tif))
                                current_geotransform = current_ds.GetGeoTransform()
                                # Compare with filepath_lc if the current raster is not filepath_lc
                                if filepath_raster != filepath_lc:
                                    print("5: Checking alignment with land cover raster...")
                                    lc_ds = gdal.Open(str(script1_outfolder / f'lc_{proj_name}_{out_res}.tif'))
                                    lc_geotransform = lc_ds.GetGeoTransform()
                                    lc_proj = lc_ds.GetProjection()
                                    lc_x_size = lc_ds.RasterXSize  # Number of columns
                                    lc_y_size = lc_ds.RasterYSize  # Number of rows
                                    if lc_geotransform != current_geotransform:
                                        print("5a: Rasters are not aligned. Attempting to re-align raster to land cover...")
                                        # GDAL Warp options to align DEM with land cover raster
                                        warp_options = gdal.WarpOptions(
                                            format='GTiff',  # Output format
                                            outputBounds=[lc_geotransform[0],  # min x
                                                          lc_geotransform[3] + lc_geotransform[5] * lc_y_size,  # min y
                                                          lc_geotransform[0] + lc_geotransform[1] * lc_x_size,  # max x
                                                          lc_geotransform[3]],  # max y
                                            xRes=lc_geotransform[1],  # x resolution (pixel width)
                                            yRes=lc_geotransform[5],  # y resolution (pixel height, should be negative)
                                            targetAlignedPixels=False,  # Align output with target pixel grid
                                            resampleAlg=gdal.GRA_Bilinear,  # Resampling method
                                            outputType=gdal.GDT_Float32,  # Output datatype
                                            dstSRS=lc_proj  # Output projection
                                        )

                                        # Perform the warp operation to align the DEM raster
                                        gdal.Warp(destNameOrDestDS=str(script1_tmp),
                                                  srcDSOrSrcDSTab=str(script1_tif),
                                                  options=warp_options)

                                        #Close all GDAL datasets
                                        current_ds = None
                                        lc_ds = None
                                        #Remove old raster and rename the adjusted raster
                                        script1_tif.unlink()
                                        script1_tmp.rename(str(script1_tif))
                                        print("5b: Raster alignment complete.")



            #SCRIPT 2
            # DCIA = a(TIA)b, where DCIA is directly connected impervious area, TIA is total impervious area, ...
            # ... according to the method of Sutherland (2000),
            # ... coefficients a and b are assigned based on land cover conditions described in Rossman et al. (2018)
            # a = 0.1 and b = 1.5 for standard connections to sewers but residential rooftops not directly connected
            # a = 0.01 and b = 2.0 for green infrastructure with sustainable urban drainage to bioswales
            dcia_coef_a = 0.1
            dcia_coef_b = 1.5
            print("6: Initializing quad processing...")

            with rasterio.open(script1_tif) as src:
                if filepath_raster == filepath_dem:
                    array = src.read(1).astype(np.float32)
                else:
                    array = src.read(1)
                array[array == src.nodata] = -9999  # Replace nodata values
                transform = src.transform
                bounds = src.bounds

                num_quads_row = int(np.ceil(src.height / row_max))
                num_quads_col = int(np.ceil(src.width / col_max))

                quad_count = 0
                #Create list to hold quads that have data
                created_quads = []
                quad_geometries = []

                if filepath_raster == filepath_dem:
                    # Set an upper threshold for valid elevation values
                    threshold = 10000

                    # Replace values larger than the threshold with NaN
                    array[array > threshold] = -9999

                    #Convert to float32
                    print("7: Calculating catchment area from DEM...")
                    cell_size_x, cell_size_y = abs(transform.a), abs(transform.e)

                    valid_cells = array != -9999  # Array where True indicates valid data
                    num_valid_cells = np.sum(valid_cells)  # Count of valid cells

                    CatchmentArea_m2 = num_valid_cells * cell_size_x * cell_size_y
                    if flag_DEM_average:
                        print("7b: Calculating average elevation from DEM...")
                        #Replace NODATA values with nan for quick calculation
                        RefWeatherLocationElevation_m = np.mean(array[array != -9999])
                        # Update the DEM average elevation in the XML configuration

                if filepath_raster == filepath_ic:
                    print("7: Calculating DCIA from impervious cover...")
                    #Get mean
                    mean_ic = np.mean(array[array != -9999])
                    #Calculate DCIA
                    dcia_dec = dcia_coef_a * mean_ic ** dcia_coef_b
                    dcia = round(dcia_dec/mean_ic, 3)
                    data_folder_config.update({"DirectlyConnectedImperviousArea_frac": dcia})


                for row in range(num_quads_row):
                    for col in range(num_quads_col):

                        row_start = row * row_max
                        row_end = min((row + 1) * row_max, src.height)
                        col_start = col * col_max
                        col_end = min((col + 1) * col_max, src.width)

                        quad_array = array[row_start:row_end, col_start:col_end]
                        if np.all(quad_array == -9999):
                            continue
                        quad_count += 1
                        quad_name = f'quad{quad_count:04d}'
                        created_quads.append(quad_name)

                        nrows, ncols = quad_array.shape

                        # Correct calculation of the lower left corner of the quadrant
                        xll_corner = bounds.left + col_start * transform.a
                        yll_corner = bounds.bottom + (src.height - row_end) * abs(transform.e)

                        if first_raster:
                            quad_geometry = box(xll_corner, yll_corner, xll_corner + ncols * transform.a,
                                                yll_corner + nrows * abs(transform.e))
                            quad_geometries.append((quad_name, quad_geometry))

                        quad_path = Path(path_project) / proj_name / f'Inputs{out_res}'/ quad_name
                        ascii_path = quad_path / script2_asc

                        create_project_folders(quad_path)

                        #Calculate area of quad array
                        cell_size_x, cell_size_y = abs(transform.a), abs(transform.e)
                        Area_m2 = cell_size_x * cell_size_y

                        # Write ASCII with header
                        with open(ascii_path, 'w') as ascii_file:
                            header_info = {
                                'ncols': ncols,
                                'nrows': nrows,
                                'xllcorner': xll_corner,
                                'yllcorner': yll_corner,
                                'cellsize': out_res,
                                'NODATA_value': -9999
                            }
                            write_ascii_header(ascii_file, header_info)
                            if script2_asc == 'dem.asc':
                                np.savetxt(ascii_file, quad_array, fmt='%.2f')
                            else:
                                np.savetxt(ascii_file, quad_array, fmt='%d')
                                #Print if it is last quad
                                if row == num_quads_row - 1 and col == num_quads_col - 1:
                                    print(f"8: Quad creation complete.")

                if first_raster:
                    gdf = gpd.GeoDataFrame(quad_geometries, columns=['quad_name', 'geometry'], crs=src.crs)
                    shapefile_path = Path(path_project) / proj_name / 'quad_extents.shp'
                    gdf.to_file(shapefile_path)
                    print("9: Quad extents shapefile created.")
                    first_raster = False

        def run_hydroplus(hydro_input_path):
            # Construct the path to HydroPlus.exe
            hydroplus_path = Path(__file__).resolve().parent.parent / 'resources' / 'HydroPlus.exe'

            # Command to run HydroPlus.exe with the given hydro_input_path
            command = [hydroplus_path, hydro_input_path]

            # Launch the command and capture the output
            process = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, text=True, bufsize=1)
            with process.stdout:
                for line in iter(process.stdout.readline, ''):
                    print(line.strip())
            process.wait()
        if flag_Hydro:
            print("=====================================\n"
                  "Hydro Project Set-Up\n"
                  "=====================================")
            hydro_input_path = Path(hydro_project_path) / proj_name / f'Inputs{str(out_res)}'
            hydro_output_path = Path(hydro_project_path) / proj_name / f'Outputs{str(out_res)}'
            create_project_folders(hydro_input_path)
            create_project_folders(hydro_output_path)

            script1_tc = script1_outfolder / f'tc_{proj_name}_{out_res}.tif'
            script1_ic = script1_outfolder / f'ic_{proj_name}_{out_res}.tif'
            script1_lc = script1_outfolder / f'lc_{proj_name}_{out_res}.tif'
            script1_dem = script1_outfolder / f'dem_{proj_name}_{out_res}.tif'

            # Load the raster data into numpy arrays for analysis
            TreeCover_array_percent, nodata_tc = read_raster(script1_tc)
            ImperviousCover_noTreeCover_array_percent, nodata_ic = read_raster(script1_ic)
            LandCover_array_class, nodata_lc = read_raster(script1_lc)

            # Filter out invalid values in the input arrays
            TreeCover_array_percent = np.clip(TreeCover_array_percent, 0, 100)
            ImperviousCover_noTreeCover_array_percent = np.clip(ImperviousCover_noTreeCover_array_percent, 0, 100)

            # Function to calculate land cover percentages for i-Tree Hydro simulations (Endreny code)
            def calculate_land_cover_percentages(ImperviousCover_noTreeCover_array_percent, TreeCover_array_percent,
                                                 LandCover_array_class, TreeCover_overImpervious_fraction):
                nodata = nodata_lc  # Use the nodata value from land cover data

                mask_mapArea_w_data = (LandCover_array_class != nodata)
                MapArea_pixel_count = np.sum(mask_mapArea_w_data)

                TreeCanopyCover_overImpervious_NonUrban_frac = 0
                TreeCanopyCover_overImpervious_Urban_frac = TreeCover_overImpervious_fraction

                # mask_mapUrbanArea_Gap is LandCover_array_class = 21-24 & (TreeCover_array_percent + ImperviousCover_noTreeCover_array_percent) < 100
                # Note: To prove array has instances of True, use print("mask_mapUrbanArea_Gap sum = {}".format(mask_mapUrbanArea_Gap.sum()))
                mask_mapUrbanArea_Gap = (LandCover_array_class >= 21) & (LandCover_array_class <= 24) & (
                            (TreeCover_array_percent + ImperviousCover_noTreeCover_array_percent) < 100)
                # mask_mapUrbanArea_NoGap is LandCover_array_class = 21-24 & (TreeCover_array_percent + ImperviousCover_noTreeCover_array_percent) >= 100
                mask_mapUrbanArea_NoGap = (LandCover_array_class >= 21) & (LandCover_array_class <= 24) & (
                            (TreeCover_array_percent + ImperviousCover_noTreeCover_array_percent) >= 100)
                # mask_mapNonUrbanArea_Gap is LandCover_array_class != 21-24 & (TreeCover_array_percent + ImperviousCover_noTreeCover_array_percent) < 100
                mask_mapNonUrbanArea_Gap = (LandCover_array_class != nodata) & (LandCover_array_class != 21) & (
                            LandCover_array_class != 22) & (LandCover_array_class != 23) & (
                                                       LandCover_array_class != 24) & ((
                                                                                                   TreeCover_array_percent + ImperviousCover_noTreeCover_array_percent) < 100)
                # mask_mapNonUrbanArea_NoGap is LandCover_array_class != 21-24 & (TreeCover_array_percent + ImperviousCover_noTreeCover_array_percent) >= 100
                mask_mapNonUrbanArea_NoGap = (LandCover_array_class != nodata) & (LandCover_array_class != 21) & (
                            LandCover_array_class != 22) & (LandCover_array_class != 23) & (
                                                         LandCover_array_class != 24) & ((
                                                                                                     TreeCover_array_percent + ImperviousCover_noTreeCover_array_percent) >= 100)

                # Calculate for both gap and no-gap areas and sum them up
                def calculate_cover_percentages(mask_mapArea_called, TreeCanopyCover_overImpervious_temp_frac):
                    # WaterCover_noTreeCover_array_percent initialized to array of zero as float, based on extent of ImperviousCover_noTreeCover_array_percent
                    WaterCover_noTreeCover_array_percent = np.zeros_like(ImperviousCover_noTreeCover_array_percent,
                                                                         dtype=np.float32)
                    SVegCover_noTreeCover_array_percent = np.zeros_like(ImperviousCover_noTreeCover_array_percent,
                                                                        dtype=np.float32)
                    SoilCover_noTreeCover_array_percent = np.zeros_like(ImperviousCover_noTreeCover_array_percent,
                                                                        dtype=np.float32)
                    # mask_waterArea_w_data defined as (LandCover_array_class > 10) & (LandCover_array_class < 20) & mask_mapArea_called
                    mask_waterArea_w_data = (LandCover_array_class > 10) & (
                                LandCover_array_class < 20) & mask_mapArea_called
                    # WaterCover_noTreeCover_array_percent[mask_waterArea_w_data] is 100 - (TC + IC)
                    WaterCover_noTreeCover_array_percent[mask_waterArea_w_data] = 100 - (
                                TreeCover_array_percent + ImperviousCover_noTreeCover_array_percent)[
                        mask_waterArea_w_data]

                    # Definition calculate_pervious_cover to find SVegCover_noTreeCover_array_percent and SoilCover_noTreeCover_array_percent values
                    def calculate_pervious_cover(LandCover_class_range, SVegCover_inGap_fraction):
                        # mask_perviousArea_w_data equals (LandCover_array_class >= LandCover_class_range[0]) & (LandCover_array_class < LandCover_class_range[1]) & mask_mapArea_called
                        mask_perviousArea_w_data = (LandCover_array_class >= LandCover_class_range[0]) & (
                                    LandCover_array_class < LandCover_class_range[1]) & mask_mapArea_called
                        SVegCover_noTreeCover_array_percent[mask_perviousArea_w_data] = SVegCover_inGap_fraction * (
                                    100 - (TreeCover_array_percent + ImperviousCover_noTreeCover_array_percent)[
                                mask_perviousArea_w_data])
                        SoilCover_noTreeCover_array_percent[mask_perviousArea_w_data] = (
                                                                                                    1 - SVegCover_inGap_fraction) * (
                                                                                                    100 - (
                                                                                                        TreeCover_array_percent + ImperviousCover_noTreeCover_array_percent)[
                                                                                                mask_perviousArea_w_data])

                    # Apply pervious cover fractions based on land cover classes
                    # NLCD Class urban 21 to 24 and < 30
                    calculate_pervious_cover((20, 30), 0.5)
                    # NLCD Class bare ground 30 to < 40
                    calculate_pervious_cover((30, 40), 0.0)
                    # NLCD Class forest 41 to 43
                    calculate_pervious_cover((40, 50), 1.0)
                    # NLCD Class agricultural 50 to < 90
                    calculate_pervious_cover((50, 90), 0.85)
                    # NLCD Class woody wetland 90 to < 95
                    calculate_pervious_cover((90, 95), 0.65)
                    # NLCD Class emergent wetland 95 to < 100
                    calculate_pervious_cover((95, 100), 0.85)

                    # return from calculate_cover_percentages sum of each cover type
                    return (
                        np.sum((1 - TreeCanopyCover_overImpervious_temp_frac) * TreeCover_array_percent[
                            mask_mapArea_called]) / MapArea_pixel_count,
                        np.sum(TreeCanopyCover_overImpervious_temp_frac * TreeCover_array_percent[
                            mask_mapArea_called]) / MapArea_pixel_count,
                        np.sum(ImperviousCover_noTreeCover_array_percent[mask_mapArea_called]) / MapArea_pixel_count,
                        np.sum(WaterCover_noTreeCover_array_percent[mask_mapArea_called]) / MapArea_pixel_count,
                        np.sum(SVegCover_noTreeCover_array_percent[mask_mapArea_called]) / MapArea_pixel_count,
                        np.sum(SoilCover_noTreeCover_array_percent[mask_mapArea_called]) / MapArea_pixel_count
                    )

                # LandCoverTypes_Urban_Gap_percent_list receives output from calculate_cover_percentages function using mask_mapUrbanArea_Gap
                # Note: Sending TreeCanopyCover_overImpervious_Urban_frac variable
                LandCoverTypes_Urban_Gap_percent_list = calculate_cover_percentages(mask_mapUrbanArea_Gap,
                                                                                    TreeCanopyCover_overImpervious_Urban_frac)
                # LandCoverTypes_Urban_NoGap_percent_list receives output from calculate_cover_percentages function using mask_mapUrbanArea_NoGap
                # Note: Sending TreeCanopyCover_overImpervious_Urban_frac variable
                LandCoverTypes_Urban_NoGap_percent_list = calculate_cover_percentages(mask_mapUrbanArea_NoGap,
                                                                                      TreeCanopyCover_overImpervious_Urban_frac)
                # LandCoverTypes_NonUrban_Gap_percent_list receives output from calculate_cover_percentages function using mask_mapNonUrbanArea_Gap
                # Note: Sending TreeCanopyCover_overImperivous_NonUrban_frac variable
                LandCoverTypes_NonUrban_Gap_percent_list = calculate_cover_percentages(mask_mapNonUrbanArea_Gap,
                                                                                       TreeCanopyCover_overImpervious_NonUrban_frac)
                # LandCoverTypes_NonUrban_NoGap_percent_list receives output from calculate_cover_percentages function using mask_mapNonUrbanArea_NoGap
                # Note: Sending TreeCanopyCover_overImpervious_NonUrban_frac variable
                LandCoverTypes_NonUrban_NoGap_percent_list = calculate_cover_percentages(mask_mapNonUrbanArea_NoGap,
                                                                                         TreeCanopyCover_overImpervious_NonUrban_frac)

                # LandCoverTypes_percent_list uses tuple to combine lc_UrbanGap + lc_UrbanNoGap + lc_NonUrbanGap + lc_NonUrbanNoGap results
                # Note: LandCoverTypes_percent_list is sum of all combined, as they were created using same denominator, MapArea_pixel_count
                LandCoverTypes_percent_list = tuple(
                    (lc_UrbanGap + lc_UrbanNoGap + lc_NonUrbanGap + lc_NonUrbanNoGap) for
                    lc_UrbanGap, lc_UrbanNoGap, lc_NonUrbanGap, lc_NonUrbanNoGap in
                    zip(LandCoverTypes_Urban_Gap_percent_list, LandCoverTypes_Urban_NoGap_percent_list,
                        LandCoverTypes_NonUrban_Gap_percent_list, LandCoverTypes_NonUrban_NoGap_percent_list))

                # return from calculate_land_cover_percentages with LandCoverTypes_percent_list
                return LandCoverTypes_percent_list


            if "TreeCanopyCover_overImpervious_frac" in data_folder_config:
                # Use the user-defined value
                TreeCover_overImpervious_fraction_CoolAir = float(data_folder_config["TreeCanopyCover_overImpervious_frac"])
                del data_folder_config["TreeCanopyCover_overImpervious_frac"]

            else:
                TreeCover_overImperviousCover_CoolAir = calculate_cool_air_frac(TreeCover_array_percent, ImperviousCover_noTreeCover_array_percent, LandCover_array_class)
            #Convert to percentage for use in calculate_land_cover_percentages, used as TreeCanopyCover_overImpervious_Urban_frac
                TreeCover_overImpervious_fraction_CoolAir = TreeCover_overImperviousCover_CoolAir/100
            print("1: Calculating urban tree cover over impervious cover fraction...")

            # Calculate land cover percentages for i-Tree Hydro simulations
            print("2: Calculating land cover percentages for i-Tree Hydro simulations...")
            results = calculate_land_cover_percentages(ImperviousCover_noTreeCover_array_percent,
                                                       TreeCover_array_percent,
                                                       LandCover_array_class, TreeCover_overImpervious_fraction_CoolAir)
            # Unpack results
            (TreeCanopyCover_overPervious_percent, TreeCanopyCover_overImpervious_percent,
             ImperviousCover_noTreeCanopy_percent, WaterCover_noTreeCanopy_percent, ShortVegCover_noTreeCanopy_percent,
             SoilCover_noTreeCanopy_percent) = results
            # TreeCanopyCover_overPervious_frac equals round((TreeCanopyCover_overPervious_percent / 100), 3)
            # Note: Equation 1st divides by 100 to create fraction, 2nd adds 0.00005 to round up to nearest 10000th, ...
            # Note: ... 3rd multiplies by 10000 to place 4 significant digits to left of decimal, 4th takes integer of number, and ...
            # Note: ... 5th and finally divides by 10000 to create fraction with 4 levels of precision and ensuring all will sum to 1.000#
            TreeCanopyCover_overPervious_frac = round((TreeCanopyCover_overPervious_percent / 100), 3)
            TreeCanopyCover_overImpervious_frac = round(TreeCanopyCover_overImpervious_percent / 100, 3)
            ImperviousCover_noTreeCanopy_frac = round((ImperviousCover_noTreeCanopy_percent / 100), 3)
            WaterCover_noTreeCanopy_frac = round((WaterCover_noTreeCanopy_percent / 100), 3)
            ShortVegCover_noTreeCanopy_frac = round((ShortVegCover_noTreeCanopy_percent / 100), 3)
            SoilCover_noTreeCanopy_frac = round((SoilCover_noTreeCanopy_percent / 100), 3)
            # TotalCover_frac equals TreeCanopyCover_overPervious_frac + TreeCanopyCover_overImpervious_frac + ImperviousCover_noTreeCanopy_frac + WaterCover_noTreeCanopy_frac + ShortVegCover_noTreeCanopy_frac + SoilCover_noTreeCanopy_frac
            TotalCover_frac = TreeCanopyCover_overPervious_frac + TreeCanopyCover_overImpervious_frac + ImperviousCover_noTreeCanopy_frac + WaterCover_noTreeCanopy_frac + ShortVegCover_noTreeCanopy_frac + SoilCover_noTreeCanopy_frac

            # If TotalCover_frac != 1.0 then enter to assign missing amount to most probable land cover in the gap
            if TotalCover_frac != 1.0:
                if ShortVegCover_noTreeCanopy_frac > 0:
                    # TotalCover_wo_ShortVeg_frac is TotalCover_frac without ShortVegCover_noTreeCanopy_frac
                    TotalCover_wo_ShortVeg_frac = TreeCanopyCover_overPervious_frac + TreeCanopyCover_overImpervious_frac + ImperviousCover_noTreeCanopy_frac + WaterCover_noTreeCanopy_frac + SoilCover_noTreeCanopy_frac
                    # ShortVegCover_noTreeCanopy_frac equals round(1.0 - TotalCover_wo_ShortVeg_frac,3)
                    ShortVegCover_noTreeCanopy_frac = round(1.0 - TotalCover_wo_ShortVeg_frac, 3)
                    # cte
                elif SoilCover_noTreeCanopy_frac > 0:
                    # TotalCover_wo_ShortVeg_frac is TotalCover_frac without SoilCover_noTreeCanopy_frac
                    TotalCover_wo_Soil_frac = TreeCanopyCover_overPervious_frac + TreeCanopyCover_overImpervious_frac + ImperviousCover_noTreeCanopy_frac + WaterCover_noTreeCanopy_frac + ShortVegCover_noTreeCanopy_frac
                    # SoilCover_noTreeCanopy_frac equals round(1.0 - TotalCover_wo_Soil_frac,3)
                    SoilCover_noTreeCanopy_frac = round(1.0 - TotalCover_wo_Soil_frac, 3)
                    # cte
                elif WaterCover_noTreeCanopy_frac > 0:
                    # TotalCover_wo_Water_frac is TotalCover_frac without WaterCover_noTreeCanopy_frac
                    TotalCover_wo_Water_frac = TreeCanopyCover_overPervious_frac + TreeCanopyCover_overImpervious_frac + ImperviousCover_noTreeCanopy_frac + ShortVegCover_noTreeCanopy_frac + SoilCover_noTreeCanopy_frac
                    # WaterCover_noTreeCanopy_frac equals round(1.0 - TotalCover_wo_Water_frac,3)
                    WaterCover_noTreeCanopy_frac = round(1.0 - TotalCover_wo_Water_frac, 3)
                    # cte
                elif WaterCover_noTreeCanopy_frac > 0:
                    # TotalCover_wo_ImpNoTree_frac is TotalCover_frac without ImperviousCover_noTreeCanopy_frac
                    TotalCover_wo_ImpNoTree_frac = TreeCanopyCover_overPervious_frac + TreeCanopyCover_overImpervious_frac + WaterCover_noTreeCanopy_frac + ShortVegCover_noTreeCanopy_frac + SoilCover_noTreeCanopy_frac
                    # ImperviousCover_noTreeCanopy_frac equals round(1.0 - TotalCover_wo_ImpNoTree_frac,3)
                    ImperviousCover_noTreeCanopy_frac = round(1.0 - TotalCover_wo_ImpNoTree_frac, 3)
                    # cte


            # Save DEM as .asc in Hydro Input folder
            print("3: Saving DEM as .asc in Hydro Input folder...")
            output_dem = hydro_input_path / 'dem.asc'
            gdal.Translate(str(output_dem), str(script1_dem), format='AAIGrid')
            # Open with numpy to round to 2nd decimal place
            with open(output_dem, 'r') as dem_file:
                # Skip the header lines (assumed to be the first 6 lines for an AAIGrid file)
                header = [next(dem_file) for _ in range(6)]

                # Read the rest of the file into a numpy array
                dem_array = np.loadtxt(dem_file)

            # Rounding the values to the 2nd decimal place
            dem_array_rounded = np.round(dem_array, 2)

            # Writing the rounded array back to the original .asc file
            with open(output_dem, 'w') as dem_file:
                # Write the header back to the file
                dem_file.writelines(header)
                # Write the rounded data to the file
                np.savetxt(dem_file, dem_array_rounded, fmt='%.2f')

            # Copy Weather and Radiation files to Hydro Input folder
            print("4: Copying Weather and Radiation files to Hydro Input folder...")
            weather_path = Path(path_weather)
            if not flag_MultipleStation:
                # Copy weather.csv and radiation.csv from weather_path
                shutil.copy(weather_path / 'Weather.csv', hydro_input_path)
                shutil.copy(weather_path / 'Radiation.csv', hydro_input_path)

                # Check if LAI.csv exists in path_weather
                lai_csv_path = weather_path / 'LAI.csv'
                if lai_csv_path.exists():
                    print("LeafOn/LeafOff days detected in LAI.csv. Extracting values...")

                    # Open the LAI.csv file and read the header
                    with open(lai_csv_path, 'r') as lai_file:
                        header = lai_file.readline().strip()  # Read the first line (header) and strip any extra whitespace

                    # Parse the header to extract StartYear_LeafOnDay and StartYear_LeafOfDay
                    header_parts = header.split(',')
                    leafon_day = int(header_parts[2].split('=')[1])
                    leafoff_day = int(header_parts[3].split('=')[1])
                else:
                    leafon_day = 1
                    leafoff_day = 365
            else:
                leafon_day = 1
                leafoff_day = 365

            # Write HydroPlusConfig.xml
            print("5: Writing HydroPlusConfig.xml...")
            hydroplus_config_path = hydro_input_path / 'HydroPlusConfig.xml'
            shutil.copy(hydroplus_config_template, hydroplus_config_path)
            out_quad_bs = str(hydro_output_path).replace('/', '\\')
            tree = ET.parse(hydroplus_config_path)
            root = tree.getroot()
            simulation_string_params = root.find('SimulationStringParams')
            if simulation_string_params is not None:
                update_xml(simulation_string_params, 'OutputFolder_Path', out_quad_bs)
                update_xml(simulation_string_params, 'Model_Selection', 'StatisticalHydro')
                if flag_MultipleStation:
                    update_xml(simulation_string_params, 'Flag_MultipleStation', '1')
                    update_xml(simulation_string_params, 'WeatherStationAttributeFile', filepath_multistation)
            simulation_numerical_params = root.find('SimulationNumericalParams')
            if simulation_numerical_params is not None:
                update_xml(simulation_numerical_params, 'CatchmentArea_m2', str(CatchmentArea_m2))
                if flag_Calibration:
                    print("5a: Calibration mode detected. Updating StartDate_YYYYMMDD and StopDate_YYYYMMDD...")
                    update_xml(simulation_numerical_params, 'StartDate_YYYYMMDD', str(calibration_start))
                    update_xml(simulation_numerical_params, 'StopDate_YYYYMMDD', str(calibration_end))
                else:
                    update_xml(simulation_numerical_params, 'StartDate_YYYYMMDD', str(simulation_start))
                    update_xml(simulation_numerical_params, 'StopDate_YYYYMMDD', str(simulation_end))
                if flag_MultipleStation:
                    with open(filepath_multistation, newline='') as file:
                        reader = csv.reader(file)
                        row_count = sum(1 for row in reader) - 1  # Subtract 1 for the header row
                        update_xml(simulation_numerical_params, 'WeatherStationNum', str(row_count))
                        update_xml(simulation_numerical_params, 'Power_InverseDistance', '2')
            data_folder = root.find('DataOrganizer/DataDrawer/DataFolder')
            if data_folder is not None:
                update_xml(data_folder, 'Area_m2', str(CatchmentArea_m2))
                update_xml(data_folder, 'TreeCanopyCover_overPervious_frac', str(TreeCanopyCover_overPervious_frac))
                update_xml(data_folder, 'TreeCanopyCover_overImpervious_frac', str(TreeCanopyCover_overImpervious_frac))
                update_xml(data_folder, 'ImperviousCover_noTreeCanopy_frac', str(ImperviousCover_noTreeCanopy_frac))
                update_xml(data_folder, 'WaterCover_noTreeCanopy_frac', str(WaterCover_noTreeCanopy_frac))
                update_xml(data_folder, 'ShortVegCover_noTreeCanopy_frac', str(ShortVegCover_noTreeCanopy_frac))
                update_xml(data_folder, 'SoilCover_noTreeCanopy_frac', str(SoilCover_noTreeCanopy_frac))
                update_xml(data_folder, 'DirectlyConnectedImperviousArea_frac', str(dcia))
                if not flag_MultipleStation:
                    update_xml(data_folder, 'LeafOnDate_JDay', str(leafon_day))
                    update_xml(data_folder, 'LeafOffDate_JDay', str(leafoff_day))
                if flag_XML_Import == 0 and data_folder_config is not None:
                    for key, value in data_folder_config.items():
                        update_xml(data_folder, key, str(value))
                if flag_Calibration:
                    # Extract Soil_FieldCapacity_m3pm3 from HydroPlusConfig.xml
                    Soil_FieldCapacity_m3pm3 = data_folder.find('Soil_FieldCapacity_m3pm3')
                    update_xml(data_folder, 'Soil_MoistureInitial_m3pm3', str(Soil_FieldCapacity_m3pm3.text))


            temperature_location_params = root.find('TemperatureLocationParams')
            if temperature_location_params is not None:
                # Delete
                root.remove(temperature_location_params)
            temperature_calculation_params = root.find('TemperatureCalculationParams')
            if temperature_calculation_params is not None:
                # Delete
                root.remove(temperature_calculation_params)
            temperature_executation_params = root.find('TemperatureExecutionParams')
            if temperature_executation_params is not None:
                # Delete
                root.remove(temperature_executation_params)

            tree.write(hydroplus_config_path)

            #Calibration
            if flag_Calibration:
                print("=====================================\n"
                      "Calibration\n"
                      "=====================================")
                # Run HydroPlus for initial values
                print("1: Running HydroPlus for initial values...")
                run_hydroplus(hydro_input_path)
                print("2: Setting up PEST_Commands...")
                # Define the path to the original and new PEST commands file
                original_pest_commands_path = Path(__file__).resolve().parent.parent / 'resources' / 'HydroPlus_PEST_Commands.csv'

                new_pest_commands_path = hydro_input_path / 'HydroPlus_PEST_Commands.csv'

                # Read the original PEST commands file
                with open(original_pest_commands_path, 'r') as file:
                    lines = file.readlines()

                # Modify the lines as per the requirements
                with open(new_pest_commands_path, 'w') as file:
                    for line in lines:
                        if line.startswith('HydroPlus_exe'):
                            hydroplus_exe_path = Path(__file__).resolve().parent.parent / 'resources' / 'HydroPlus.exe'
                            line = f'HydroPlus_exe,{hydroplus_exe_path}\n'
                        elif line.startswith('HydroPlus_input'):
                            line = f'HydroPlus_input,{hydro_input_path}\n'
                        elif line.startswith('PEST_date_YYYYMMDDHHMM,SimulationNumericalParams/StartDate_YYYYMMDD'):
                            line = f'PEST_date_YYYYMMDDHHMM,SimulationNumericalParams/StartDate_YYYYMMDD,{calibration_start}\n'
                        elif line.startswith('PEST_date_YYYYMMDDHHMM,SimulationNumericalParams/StopDate_YYYYMMDD'):
                            line = f'PEST_date_YYYYMMDDHHMM,SimulationNumericalParams/StopDate_YYYYMMDD,{calibration_end}\n'
                        elif line.startswith('PEST_observation_file') and 'TS_first_StorageDeficit_VadoseZone.csv' in line:
                            line = f'PEST_observation_file,{hydro_input_path / "TS_first_StorageDeficit_VadoseZone.csv"}\n'
                        elif line.startswith('PEST_observation_file') and 'TS_Last_StorageDeficit_VadoseZone.csv' in line:
                            line = f'PEST_observation_file,{hydro_input_path / "TS_Last_StorageDeficit_VadoseZone.csv"}\n'
                        elif line.startswith('HydroPlus_prediction'):
                            line = f'HydroPlus_prediction,{hydro_output_path / "Summary_Budget_Water_mm.csv"}\n'

                        file.write(line)

                if flag_Vadose_auto:
                    print("3: Calculating balanced vadose zone storage deficit...")
                    # Read the summary budget file
                    summary_budget_path = hydro_output_path / 'TimeSeries_PerviousVariables_Catchment_m.csv'
                    storage_deficit_values = []

                    # Read the CSV file
                    with open(summary_budget_path, 'r') as csvfile:
                        reader = csv.reader(csvfile)
                        header = next(reader)
                        # Find the index of the StorageDeficit_VadoseZone_m column
                        storage_deficit_index = header.index('StorageDeficit_VadoseZone_m')

                        for row in reader:
                            storage_deficit_values.append(float(row[storage_deficit_index]))

                    # Remove the first 720 and last 720 values
                    trimmed_values = storage_deficit_values[720:-720]

                    # Calculate the average
                    avg_value = np.mean(trimmed_values)
                    #Round to 3 decimal places
                    avg_value_mm = round(avg_value * 1000, 2)
                    # Create TS_first_StorageDeficit_VadoseZone.csv
                    first_csv_path = hydro_input_path / 'TS_first_StorageDeficit_VadoseZone.csv'
                    with open(first_csv_path, 'w', newline='') as csvfile:
                        writer = csv.writer(csvfile)
                        writer.writerow(['Observation value:'])
                        writer.writerow(['TS_first_StorageDeficit_VadoseZone', avg_value_mm, 'mm'])

                    # Create TS_Last_StorageDeficit_VadoseZone.csv
                    last_csv_path = hydro_input_path /'TS_Last_StorageDeficit_VadoseZone.csv'
                    with open(last_csv_path, 'w', newline='') as csvfile:
                        writer = csv.writer(csvfile)
                        writer.writerow(['Observation value:'])
                        writer.writerow(['TS_Last_StorageDeficit_VadoseZone', avg_value_mm, 'mm'])
                else:
                    # Create TS_first_StorageDeficit_VadoseZone.csv
                    first_csv_path = hydro_input_path / 'TS_first_StorageDeficit_VadoseZone.csv'
                    with open(first_csv_path, 'w', newline='') as csvfile:
                        writer = csv.writer(csvfile)
                        writer.writerow(['Observation value:'])
                        writer.writerow(['TS_first_StorageDeficit_VadoseZone', first_vadose, 'mm'])

                    # Create TS_Last_StorageDeficit_VadoseZone.csv
                    last_csv_path = hydro_input_path / 'TS_Last_StorageDeficit_VadoseZone.csv'
                    with open(last_csv_path, 'w', newline='') as csvfile:
                        writer = csv.writer(csvfile)
                        writer.writerow(['Observation value:'])
                        writer.writerow(['TS_Last_StorageDeficit_VadoseZone', last_vadose, 'mm'])
                print("3: Launching PEST set-up...")
                # Path to the z_b_CalibrationPEST_iTHydroPlus.py script
                calibration_script_path = Path(__file__).resolve().parent / 'z_b_CalibrationPEST_iTHydroPlus.py'

                # Command to run the calibration script
                command = [sys.executable, calibration_script_path, hydro_input_path]

                # Launch the command and capture the output
                process = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, text=True,
                                           bufsize=1)
                with process.stdout:
                    for line in iter(process.stdout.readline, ''):
                        print(line.strip())
                process.wait()

                print("4: Launching PEST...")
                # Construct the path to i64pest.exe
                i64pest_path = Path(__file__).resolve().parent.parent / 'resources' / 'i64pest.exe'
                control_file = 'HydroPlus_PEST_ControlFile.pst'

                # Change the working directory to hydro_input_path
                process = subprocess.Popen([i64pest_path, control_file], cwd=hydro_input_path, stdout=subprocess.PIPE,
                                           stderr=subprocess.STDOUT, text=True, bufsize=1)

                # Capture and print the output
                with process.stdout:
                    for line in iter(process.stdout.readline, ''):
                        print(line.strip())
                process.wait()


        else:
            if "TreeCanopyCover_overImpervious_frac" in data_folder_config:
                # Use the user-defined value
                TreeCover_overImpervious_fraction_CoolAir = float(data_folder_config["TreeCanopyCover_overImpervious_frac"])
                del data_folder_config["TreeCanopyCover_overImpervious_frac"]
            else:
                # Still need to check for tc_oic and calculate if it does not exist
                script1_tc = script1_outfolder / f'tc_{proj_name}_{out_res}.tif'
                script1_ic = script1_outfolder / f'ic_{proj_name}_{out_res}.tif'
                script1_lc = script1_outfolder / f'lc_{proj_name}_{out_res}.tif'
                script1_dem = script1_outfolder / f'dem_{proj_name}_{out_res}.tif'
                # Load the raster data into numpy arrays for analysis
                TreeCover_array_percent, nodata_tc = read_raster(script1_tc)
                ImperviousCover_noTreeCover_array_percent, nodata_ic = read_raster(script1_ic)
                LandCover_array_class, nodata_lc = read_raster(script1_lc)

                TreeCover_overImperviousCover_CoolAir = calculate_cool_air_frac(TreeCover_array_percent, ImperviousCover_noTreeCover_array_percent, LandCover_array_class)
                TreeCover_overImpervious_fraction_CoolAir = TreeCover_overImperviousCover_CoolAir/100



        print("==========================\n"
              "Quad Processing\n"
              "==========================\n"
              "For each quad, the following steps are performed:\n"
                "1. Copy weather and radiation files to quad folder\n"
                "2. Remove tree cover and impervious cover over water\n"
                "3. Create HydroPlusConfig.xml\n"
                "4. Adjust calibrated parameters, if applicable\n"
                "5. Create alternate scenarios, if applicable\n"
                "6. Add quad folder to batch list")
        #Initialize header info to write to HydroPlusConfig
        header_info = {}
        # Total number of quads to process
        total_quads = len(created_quads)

        # Progress step for each 10%
        progress_step = total_quads / 10

        # Counter to track progress
        progress_count = 0
        #List for batch
        all_quads = []
        # Base scenario quads
        from pathlib import Path

        base_quad_paths = [Path(path_project) / proj_name / f'Inputs{out_res}' / quad_name for quad_name in created_quads]
        all_quads.extend(base_quad_paths)
        weather_path = Path(path_weather)

        # Check if LAI.csv exists in path_weather
        lai_csv_path = weather_path / 'LAI.csv'
        if lai_csv_path.exists():

            # Open the LAI.csv file and read the header
            with open(lai_csv_path, 'r') as lai_file:
                header = lai_file.readline().strip()  # Read the first line (header) and strip any extra whitespace

            # Parse the header to extract StartYear_LeafOnDay and StartYear_LeafOfDay
            header_parts = header.split(',')
            leafon_day = int(header_parts[2].split('=')[1])
            leafoff_day = int(header_parts[3].split('=')[1])
        else:
            leafon_day = 1
            leafoff_day = 365

        for index, quad_name in enumerate(created_quads):
            quad_path = Path(path_project) / proj_name / f'Inputs{out_res}' / quad_name
            # Update progress counter
            progress_count += 1

            # Check if progress_count has reached the next 10% milestone
            if progress_count >= progress_step:
                # Calculate the completed percentage
                percent_complete = (index + 1) / total_quads * 100
                # Send update to console via the update_console method or similar
                print(f"Quad Processing: {percent_complete:.1f}% complete. Processing {quad_name}...")
                # Reset progress count after each message
                progress_count = 0
            if not flag_MultipleStation:
                # Copy weather.csv and radiation.csv from weather_path
                shutil.copy(weather_path / 'Weather.csv', quad_path)
                shutil.copy(weather_path / 'Radiation.csv', quad_path)

            #Open DEM and convert any 21474836.00 to -9999
            dem_path = quad_path / 'dem.asc'
            header = []
            with open(dem_path, 'r') as file:
                for _ in range(6):  # Assuming the header is 6 lines
                    header.append(next(file))
                dem_data = np.loadtxt(file)

            dem_noval = np.where(dem_data == 21474836.00)
            dem_data[dem_noval] = -9999
            with open(dem_path, 'w') as file:
                file.write(''.join(header))
                np.savetxt(file, dem_data, fmt='%.2f')

            # Read and store header for 'landcover.asc'
            land_cover_path = quad_path / 'landcover.asc'
            with open(land_cover_path, 'r') as file:
                lc_header = [next(file) for _ in range(6)]
                for line in lc_header:
                    key, value = line.split()
                    header_info[key] = value
                land_cover_data = np.loadtxt(file)

            # Find locations where land cover is 11
            lc11_locations = np.where(land_cover_data == 11)
            lc_urban = np.where(((land_cover_data < 21) | (land_cover_data > 24)) & (land_cover_data != -9999))
            #print("lc_11 locations=", lc11_locations)

            # Process 'treecover.asc' and 'impervious.asc'
            for file_name in ['treecover.asc', 'imperviouscover.asc']:
                file_path = quad_path / file_name
                # Read and store the original header
                header = []
                with open(file_path, 'r') as file:
                    for _ in range(6):  # Assuming the header is 6 lines
                        header.append(next(file))
                    data = np.loadtxt(file)  # Load the data after the header
                # Update values to 0 where land cover is 11
                data[lc11_locations] = 0
                if file_name == 'imperviouscover.asc':
                    # Update values to 0 where land cover is not 21-24
                    data[lc_urban] = 0
                # Write the updated data back to the file
                with open(file_path, 'w') as file:
                    # Join the header into a single string and write it
                    file.write(''.join(header))
                    # Save the updated data array
                    np.savetxt(file, data, fmt='%d')


            #Alternate Scenarios
            scenario_quads = []
            if scenarios is not None:
                lc_path = quad_path / 'landcover.asc'
                tc_path = quad_path / 'treecover.asc'
                ic_path = quad_path / 'imperviouscover.asc'

                def read_ascii_file(file_path):
                    with open(file_path, 'r') as file:
                        data = np.loadtxt(file, skiprows=6)
                    return data
                def write_ascii_file(file_path, data, metadata):
                    with open(file_path, 'w') as file:
                        file.write(metadata)
                        np.savetxt(file, data, fmt='%d')
                def apply_where_conditions(data, conditions):
                    condition_map = {
                        "equals": lambda x, y: x == y,
                        "is greater than": lambda x, y: x > y,
                        "is less than": lambda x, y: x < y,
                        "is greater than or equal to": lambda x, y: x >= y,
                        "is less than or equal to": lambda x, y: x <= y,
                        "is not equal to": lambda x, y: x != y
                    }

                    no_data_value = -9999
                    combined_mask = np.ones(data.shape, dtype=bool)

                    for i, cond in enumerate(conditions):
                        field_condition = condition_map[cond["condition"]]
                        value = float(cond["value"])

                        mask = field_condition(data, value)
                        mask[data == no_data_value] = False

                        #print(f"Condition: {cond['condition']} {value}, Mask sum: {np.sum(mask)}")

                        if i == 0:
                            combined_mask = mask
                        else:
                            operator = cond.get("operator", "")
                            if operator == "OR":
                                combined_mask = combined_mask | mask
                            else:  # Default to AND if no operator or explicitly AND
                                combined_mask = combined_mask & mask

                        #print(f"Applied condition: {cond}, resulting mask sum: {np.sum(combined_mask)}")

                    return combined_mask
                def apply_action(data, mask, actions):
                    no_data_value = -9999

                    for action in actions:
                        operation = action["operation"]
                        value = float(action["value"])

                        #print(f"Applying action: {action} on {np.sum(mask)} cells")

                        affected_indices = np.where(mask & (data != no_data_value))

                        if operation == "add":
                            data[affected_indices] += value
                        elif operation == "subtract":
                            data[affected_indices] -= value
                        elif operation == "multiply":
                            data[affected_indices] *= value
                        elif operation == "divide":
                            data[affected_indices] /= value
                        elif operation == "set equal to":
                            data[affected_indices] = value

                        # Clamp values to the range [0, 100]
                        data[affected_indices] = np.clip(data[affected_indices], 0, 100)

                        #print(f"Finished applying action: {operation} with value: {value}")
                        #print(f"Data after applying action: {data[affected_indices]}")
                def process_scenarios(scenarios, tc_path, lc_path, ic_path):
                    # Read the ASCII files and metadata once
                    with open(tc_path, 'r') as file:
                        tc_metadata = ''.join([file.readline() for _ in range(6)])

                    with open(lc_path, 'r') as file:
                        lc_metadata = ''.join([file.readline() for _ in range(6)])

                    with open(ic_path, 'r') as file:
                        ic_metadata = ''.join([file.readline() for _ in range(6)])

                    for scenario in scenarios:
                        scenario_name = scenario["name"]
                        where_conditions = scenario["where"]
                        actions = scenario["action"]

                        # Reload the original data for each scenario
                        tc_data = read_ascii_file(tc_path)
                        lc_data = read_ascii_file(lc_path)
                        ic_data = read_ascii_file(ic_path)

                        data_map = {
                            "tree cover": (tc_data, tc_metadata, tc_path),
                            "land cover": (lc_data, lc_metadata, lc_path),
                            "impervious cover": (ic_data, ic_metadata, ic_path)
                        }

                        # Apply where conditions to the corresponding field first
                        combined_mask = np.ones(tc_data.shape, dtype=bool)
                        for cond in where_conditions:
                            where_field = cond["field"]
                            data, metadata, path = data_map[where_field]
                            #print(f"Applying conditions to field: {where_field}")
                            mask = apply_where_conditions(data, [cond])
                            combined_mask &= mask

                        # Apply actions based on the combined mask to the action field
                        for action in actions:
                            action_field = action["field"]
                            action_data, action_metadata, action_path = data_map[action_field]
                            #print(f"Applying actions to field: {action_field}")
                            apply_action(action_data, combined_mask, [action])

                        quad_path_scenario = Path(path_project) / proj_name / scenario_name / f'Inputs{out_res}' / quad_name
                        if not quad_path_scenario.exists():
                            quad_path_scenario.mkdir(parents=True, exist_ok=True)

                        # Write the modified data back to ASCII files
                        write_ascii_file(quad_path_scenario / 'treecover.asc', tc_data, tc_metadata)
                        write_ascii_file(quad_path_scenario / 'landcover.asc', lc_data, lc_metadata)
                        write_ascii_file(quad_path_scenario / 'imperviouscover.asc', ic_data, ic_metadata)

                        # Copy other files to quad folder
                        for file_name in ['dem.asc', 'Weather.csv', 'Radiation.csv']:
                            original_file_path = Path(quad_path) / file_name
                            shutil.copy(original_file_path, quad_path_scenario / file_name)
                        #Then, edit the tree and impervious cover files to comply with water cover and non-urban cover
                        # Process 'treecover.asc' and 'impervious.asc'
                        for file_name in ['treecover.asc', 'imperviouscover.asc']:
                            file_path = quad_path_scenario / file_name
                            # Read and store the original header
                            header = []
                            with open(file_path, 'r') as file:
                                for _ in range(6):  # Assuming the header is 6 lines
                                    header.append(next(file))
                                data = np.loadtxt(file)  # Load the data after the header
                            # Update values to 0 where land cover is 11
                            data[lc11_locations] = 0
                            if file_name == 'imperviouscover.asc':
                                # Update values to 0 where land cover is not 21-24
                                data[lc_urban] = 0
                            # Write the updated data back to the file
                            with open(file_path, 'w') as file:
                                # Join the header into a single string and write it
                                file.write(''.join(header))
                                # Save the updated data array
                                np.savetxt(file, data, fmt='%d')

                        scenario_quads.append(quad_path_scenario)

                    if scenario_quads:
                        all_quads.extend(scenario_quads)
                # Process scenarios for the current quad
                process_scenarios(scenarios, tc_path, lc_path, ic_path)

            if flag_XML_Import:
                hydroplus_config_template = Path(path_trunk) / proj_name / 'HydroPlusConfig_Project_Template.xml'
            else:
                hydroplus_config_template_out = Path(path_trunk) / proj_name / 'HydroPlusConfig_Project_Template.xml'
            # Open HydroPlusConfig.xml
            hydroplus_config_path = Path(path_project) / proj_name / f'Inputs{out_res}' / quad_name / 'HydroPlusConfig.xml'
            # Copy the template HydroPlusConfig.xml to the quad folder
            shutil.copy(hydroplus_config_template, hydroplus_config_path)
            out_quad = Path(path_project) / proj_name / f'Outputs{out_res}' / quad_name
            #Create output quad folder
            create_project_folders(out_quad)
            out_quad_bs = str(out_quad).replace('/', '\\')
            tree = ET.parse(hydroplus_config_path)
            root = tree.getroot()
            simulation_string_params = root.find('SimulationStringParams')
            if simulation_string_params is not None:
                update_xml(simulation_string_params, 'OutputFolder_Path', out_quad_bs)
                update_xml(simulation_string_params, 'Model_Selection', 'SpatialTemperatureHydro')
                if flag_MultipleStation:
                    update_xml(simulation_string_params, 'Flag_MultipleStation', '1')
                    update_xml(simulation_string_params, 'WeatherStationAttributeFile', filepath_multistation)
                    if flag_HottestDaySimulated:
                        update_xml(simulation_string_params, 'Flag_HottestDaySimulated', '1')
                    else:
                        update_xml(simulation_string_params, 'Flag_HottestDaySimulated', '0')
                if flag_AnthropogenicHeat:
                    update_xml(simulation_string_params, 'Flag_CoolAir_AnthropogenicHeat', '1')
                else:
                    update_xml(simulation_string_params, 'Flag_CoolAir_AnthropogenicHeat', '0')
            simulation_numerical_params = root.find('SimulationNumericalParams')
            if simulation_numerical_params is not None:
                update_xml(simulation_numerical_params, 'StartDate_YYYYMMDD', str(simulation_start))
                update_xml(simulation_numerical_params, 'StopDate_YYYYMMDD', str(simulation_end))
                update_xml(simulation_numerical_params, 'CatchmentArea_m2', str(CatchmentArea_m2))
                if flag_MultipleStation:
                    with open(filepath_multistation, newline='') as file:
                        reader = csv.reader(file)
                        row_count = sum(1 for row in reader) - 1  # Subtract 1 for the header row
                        update_xml(simulation_numerical_params, 'WeatherStationNum', str(row_count))
                        update_xml(simulation_numerical_params, 'Power_InverseDistance', '2')
            data_folder = root.find('DataOrganizer/DataDrawer/DataFolder')
            if data_folder is not None:
                update_xml(data_folder, 'Area_m2', str(Area_m2))
                update_xml(data_folder, 'LeafOnDate_JDay', str(leafon_day))
                update_xml(data_folder, 'LeafOffDate_JDay', str(leafoff_day))
                update_xml(data_folder, 'TreeCanopyCover_overImpervious_frac', str(TreeCover_overImpervious_fraction_CoolAir))
                if flag_XML_Import == 0 and data_folder_config is not None:
                    for key, value in data_folder_config.items():
                        update_xml(data_folder, key, str(value))
                Soil_FieldCapacity_m3pm3 = data_folder.find('Soil_FieldCapacity_m3pm3')
                Soil_WiltingPoint_m3pm3 = data_folder.find('Soil_WiltingPoint_m3pm3')
                #Find halfway between FieldCapacity and WiltingPoint
                Soil_MoistureInitial_m3pm3 = (float(Soil_FieldCapacity_m3pm3.text) + float(Soil_WiltingPoint_m3pm3.text)) / 2
                update_xml(data_folder, 'Soil_MoistureInitial_m3pm3', str(Soil_MoistureInitial_m3pm3))


                if flag_Calibration:
                    hydro_config_path = hydro_input_path / 'HydroPlusConfig.xml'

                    # Read the HydroPlusConfig.xml
                    hydro_tree = ET.parse(hydro_config_path)
                    hydro_root = hydro_tree.getroot()

                    # Extract the required values with the proper structure
                    vadose_transmissivity = hydro_root.find(
                        'DataOrganizer/DataDrawer/DataFolder/VadoseZone_Transmissivity_Max_m2ph').text
                    ksat_exp_decay = hydro_root.find(
                        'DataOrganizer/DataDrawer/DataFolder/Parameter_m_KsatExpDecay').text
                    discharge_subsurface_initial = hydro_root.find(
                        'DataOrganizer/DataDrawer/DataFolder/Discharge_Subsurface_Initial_mph').text
                    # Update the data_folder structure in the xml we are updating
                    update_xml(data_folder, 'VadoseZone_Transmissivity_Max_m2ph', vadose_transmissivity)
                    update_xml(data_folder, 'Parameter_m_KsatExpDecay', ksat_exp_decay)
                    update_xml(data_folder, 'Discharge_Subsurface_Initial_mph', discharge_subsurface_initial)
            temperature_location_params = root.find('TemperatureLocationParams')
            #if temperature_location_params is not None:
            #fill in nCols, nRows, NODATA, cellsize, xllcorner, and yllcorner
            if temperature_location_params is not None:
                update_xml(temperature_location_params, 'nCols', header_info['ncols'])
                update_xml(temperature_location_params, 'nRows', header_info['nrows'])
                update_xml(temperature_location_params, 'NODATA', '-9999')
                update_xml(temperature_location_params, 'cellsize', out_res)
                update_xml(temperature_location_params, 'xllcorner', header_info['xllcorner'])
                update_xml(temperature_location_params, 'yllcorner', header_info['yllcorner'])
            temperature_calculation_params = root.find('TemperatureCalculationParams')
            if temperature_calculation_params is not None:
                if not flag_XML_Import and temperature_calculation_config is not None:
                    for key, value in temperature_calculation_config.items():
                        update_xml(temperature_calculation_params, key, str(value))
            temperature_executation_params = root.find('TemperatureExecutionParams')
            if temperature_executation_params is not None:
                if flag_DEM_average:
                    update_xml(temperature_executation_params, 'RefWeatherLocationElevation_m', str(RefWeatherLocationElevation_m))
                if not flag_XML_Import and temperature_execution_config is not None:
                        for key, value in temperature_execution_config.items():
                            update_xml(temperature_executation_params, key, str(value))
                if not flag_BlockGroup:
                    delete_xml(temperature_executation_params, 'PrintBlockGroupRange0')
                    update_xml(temperature_executation_params, 'PrintBlockGroupDailyAndHourly', 'False')
                else:
                    update_xml(temperature_executation_params, 'PrintBlockGroupRange0', census_fids)
                    update_xml(temperature_executation_params, 'PrintBlockGroupDailyAndHourly', 'True')
            tree.write(hydroplus_config_path)
            if not Path(hydroplus_config_template_out).exists() and not flag_XML_Import:
                #Copy HydroPlusConfig.xml to HydroPlusConfig_Project_Template.xml
                shutil.copy(hydroplus_config_path, hydroplus_config_template_out)

            if scenarios is not None:
                for scenario in scenarios:
                    scenario_name = scenario["name"]
                    out_quad_scenario = Path(path_project) / proj_name / scenario_name / f'Outputs{out_res}' / quad_name
                    create_project_folders(out_quad_scenario)
                    out_quad_bs_scenario = str(out_quad_scenario).replace('/', '\\')
                    hydroplus_config_path_scenario = Path(path_project) / proj_name / scenario_name / f'Inputs{out_res}' / quad_name / 'HydroPlusConfig.xml'
                    shutil.copy(hydroplus_config_path, hydroplus_config_path_scenario)
                    tree = ET.parse(hydroplus_config_path_scenario)
                    root = tree.getroot()
                    simulation_string_params = root.find('SimulationStringParams')
                    if simulation_string_params is not None:
                        update_xml(simulation_string_params, 'OutputFolder_Path', out_quad_bs_scenario)
                    tree.write(hydroplus_config_path_scenario)

        # Write quad_list to batch files
        if scenarios is not None:
            for scenario in scenarios:
                scenario_name = scenario["name"]
                scenario_output_folder = Path(path_project) / proj_name / scenario_name / f'Outputs{out_res}'
                output_folders_list.append(scenario_output_folder)
        number_quads = len(all_quads)
        batch_prefix = f'{proj_name}_{out_res}'
        #print(f"10: {number_quads} total quads created for all scenarios.")
        quads_per_batch = number_quads // flag_NumberofBatches
        remaining_quads = number_quads % flag_NumberofBatches
        print("5: Creating batch files...")

        for i in range(1, flag_NumberofBatches + 1):
            start_idx = (i - 1) * quads_per_batch
            if i == flag_NumberofBatches:
                end_idx = start_idx + quads_per_batch + remaining_quads
            else:
                end_idx = start_idx + quads_per_batch

            quads_list = all_quads[start_idx:end_idx]
            create_batch_file(Path(path_project) / proj_name, i, quads_list, batch_prefix)
            batch_files.append(Path(path_project) / proj_name / f'{proj_name}_{out_res}-{i}.bat')

            #Write batch file
    for batch_file in batch_files:
        print(f"{batch_file}")
    for output_folder in output_folders_list:
        print(f"output_folder = {output_folder}")



if __name__ == "__main__":
    print("Starting Cool Air Prep script p_itca_01_prep.py...")
    parser = argparse.ArgumentParser(description='Process project-specific configuration.')
    parser.add_argument('--proj_name', type=str, required=True, help='Project name')
    parser.add_argument('--config_filename', type=str, required=True, help='Full path to the configuration JSON file')
    args = parser.parse_args()
    main(args)




