import os
import glob
import datetime
from osgeo import gdal
import subprocess 
import rasterio
import rasterio.crs
from rasterio.enums import Resampling
from rasterio.merge import merge
import argparse
import json
import numpy as np
from pathlib import Path
gdal.UseExceptions()


def adjust_first_value(ascii_file):
    try:
        with open(ascii_file, 'r') as file:
            lines = file.readlines()

        # Extract the header (first 6 lines)
        header = lines[:6]

        # Extract the data part
        data_lines = lines[6:]
        data = [list(map(float, line.split())) for line in data_lines]

        # Calculate the average of surrounding values of the first data cell (0, 0)
        if len(data) > 1 and len(data[0]) > 1:
            neighbors = [
                data[0][1],     # right
                data[1][0],     # below
                data[1][1]      # below-right
            ]
            data[0][0] = np.mean(neighbors)

        # Write the modified data back to the file
        with open(ascii_file, 'w') as file:
            file.writelines(header)
            for row in data:
                file.write(' '.join(map(str, row)) + '\n')

    except ValueError as e:
        print(f"Error processing file {ascii_file}: {e}")

def ascii_to_tif(ascii_file, tif_file, output_epsg):
    # Ensure paths are strings
    ascii_file = str(ascii_file)
    tif_file = str(tif_file)

    # Adjust the first value of the ASCII file
    adjust_first_value(ascii_file)

    # Convert the adjusted ASCII to TIF
    src_ds = gdal.Open(ascii_file)
    gdal.Translate(tif_file, src_ds, options=f"-a_srs EPSG:{output_epsg} -a_nodata -9999")

def mosaic_tifs(input_tifs, output_tif):
    # Initialize the list to hold the rasterio Dataset objects
    src_files_to_mosaic = []

    # Open each tif file and append the rasterio Dataset object to src_files_to_mosaic
    for tif in input_tifs:
        src = rasterio.open(tif)
        src_files_to_mosaic.append(src)

    if len(src_files_to_mosaic) == 0:
        return

    # Merge files
    try:
        mosaic, out_transform = merge(src_files_to_mosaic)
    except Exception as e:
        print(f"Error during merging: {e}")
        return

    # Set values less than 0 to -9999
    mosaic[mosaic < 0] = -9999

    # Use the metadata from one of the source files for the output metadata
    out_meta = src_files_to_mosaic[0].meta.copy()
    out_meta.update({
        "driver": "GTiff",
        "height": mosaic.shape[1],
        "width": mosaic.shape[2],
        "transform": out_transform,
        "nodata": -9999,
        # Update other metadata if necessary
    })

    # Write the clipped mosaic to a new file
    with rasterio.open(output_tif, "w", **out_meta) as dest:
        dest.write(mosaic)

    # Close the rasterio dataset objects
    for src in src_files_to_mosaic:
        src.close()


def k_to_f(input_tif, output_folder, current_day_str, temp_type):
    output_tif_path = output_folder / f"{temp_type}_F_{current_day_str}.tif"
    with rasterio.open(input_tif) as src:
        data = src.read(1)  # Read the first band
        nodata_value = -9999

        # Create a mask for valid data
        valid_mask = data != nodata_value

        # Initialize the output array with the nodata value
        fahrenheit_data = np.full_like(data, nodata_value, dtype=rasterio.float32)

        # Apply the Kelvin to Fahrenheit conversion only to valid data
        print("Converting Kelvin to Fahrenheit for valid data...")
        fahrenheit_data[valid_mask] = (data[valid_mask] - 273.15) * 9/5 + 32

        profile = src.profile
        profile.update(dtype=rasterio.float32)

        with rasterio.open(output_tif_path, 'w', **profile) as dst:
            dst.write(fahrenheit_data, 1)  # Write the Fahrenheit data to the first band

def k_to_c(input_tif, output_folder, current_day_str, temp_type):
    output_tif_path = output_folder / f"{temp_type}_C_{current_day_str}.tif"
    with rasterio.open(input_tif) as src:
        data = src.read(1)
        nodata_value = -9999

        valid_mask = data != nodata_value
        celsius_data = np.full_like(data, nodata_value, dtype=rasterio.float32)
        celsius_data[valid_mask] = data[valid_mask] - 273.15

        profile = src.profile
        profile.update(dtype=rasterio.float32)

        with rasterio.open(output_tif_path, 'w', **profile) as dst:
            dst.write(celsius_data, 1)  # Write the Celsius data to the first band


def main(args):
    output_folder = Path(args.output_folder)
    mosaic_folder = output_folder / 'mosaic_map'
    if not mosaic_folder.exists():
        mosaic_folder.mkdir(parents = True, exist_ok = True)

    if args.startday and args.endday:
        StartDay = args.startday
        EndDay = args.endday
        try:
            start_datetime = datetime.datetime.strptime(StartDay, "%Y%m%d%H")
            end_datetime = datetime.datetime.strptime(EndDay, "%Y%m%d%H")
        except ValueError:
            StartDay = StartDay + "00"
            EndDay = EndDay + "23"
            start_datetime = datetime.datetime.strptime(StartDay, "%Y%m%d%H")
            end_datetime = datetime.datetime.strptime(EndDay, "%Y%m%d%H")
    else:
        #TODO: Auto search for start and end based on a full quad folder
        print("Start and end day not valid. Exiting function.", flush=True)

    print(f"Mosaic Destination: {mosaic_folder}\n"
          f"Start Day: {start_datetime}\n"
          f"End Day: {end_datetime}\n"
          f"Output EPSG: {args.output_epsg}\n"
          f"Temperature Units: {args.temp_units}\n")
    print("=====================================================\n"
          "Processing Mosaic Maps...\n"
          "=====================================================\n")
    #Addition of different output types to accomodate Heat Index.
    file_types = ['TAir_K', 'HeatIndex_K']
    for file_type in file_types:
        if file_type == 'TAir_K':
            print("Processing Temperature Maps...")
            temp_type = 'TAir'
        elif file_type == 'HeatIndex_K':
            print("Processing Heat Index Maps...")
            temp_type = 'HeatIndex'
        current_datetime = start_datetime
        while current_datetime <= end_datetime:
            current_day_str = current_datetime.strftime("%Y%m%d%H")
            print(f"1: Collecting ASCIIs for {current_day_str}...")
            ascii_files = glob.glob(str(Path(output_folder) / '*' / f"{file_type}_{current_day_str}.asc"))

            tif_files = []
            file_counter = 1

            for ascii_file in ascii_files:
                tif_file = Path(mosaic_folder) / f"{file_type}_{current_day_str}_{file_counter}.tif"
                ascii_to_tif(ascii_file, tif_file, args.output_epsg)
                tif_files.append(tif_file)
                file_counter += 1
            print("1. Number of ASCIIs found: ", len(ascii_files))
            output_tif = Path(mosaic_folder) / f"{temp_type}_mosaic_{current_day_str}.tif"
            print(f"2: Mosaicing TIFs for {current_day_str}...")
            if tif_files:
                mosaic_tifs(tif_files, output_tif)
                if args.temp_units.get('fahrenheit', True):
                    k_to_f(output_tif, mosaic_folder, current_day_str, temp_type)
                    print("3a: Converting Kelvin to Fahrenheit...")
                if args.temp_units.get('celsius', True):
                    k_to_c(output_tif, mosaic_folder, current_day_str, temp_type)
                    print("3b: Converting Kelvin to Celsius...")

            print("4: Removing temporary TIF files...")
            for tif_file in tif_files:
                Path(tif_file).unlink()

            current_datetime += datetime.timedelta(hours=1)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Automate mosaic processing of output maps.")
    parser.add_argument('--output_folder', type=str, required=True,
                        help="The folder containing the outputs to process.")
    parser.add_argument('--output_epsg', type=int, required=True,
                        help="The EPSG code for the output coordinate system.")
    parser.add_argument('--startday', type=str, help="Start day for processing (optional).")
    parser.add_argument('--endday', type=str, help="End day for processing (optional).")
    parser.add_argument('--temp_units', type=json.loads,  # Use json.loads to parse the json string back to a dictionary
                        required=True, help="JSON string representing temperature units preferences.")
    args = parser.parse_args()
    main(args)